#!/usr/bin/env python3

import unittest

import pytorch_lightning as pl
from pytorch.recipes.rec.datamodules.random_rec_datamodule import RandomRecDataModule
from pytorch.recipes.rec.modules.sparsenn_classification import (
    SparseNNBinaryClassification,
)
from torchrec import EmbeddingBagCollection
from torchrec.modules.embedding_configs import EmbeddingBagConfig


class TestSparseNNClassification(unittest.TestCase):
    def test_train_model(self) -> None:
        num_embeddings = 100
        embedding_dim = 10

        eb1_config = EmbeddingBagConfig(
            name="t1",
            embedding_dim=embedding_dim,
            num_embeddings=num_embeddings,
            feature_names=["f1", "f3"],
        )
        eb2_config = EmbeddingBagConfig(
            name="t2",
            embedding_dim=embedding_dim,
            num_embeddings=num_embeddings,
            feature_names=["f2"],
        )
        ebc = EmbeddingBagCollection(tables=[eb1_config, eb2_config])

        model = SparseNNBinaryClassification(ebc)
        datamodule = RandomRecDataModule()

        trainer = pl.Trainer(
            max_epochs=3,
            limit_train_batches=100,
            limit_val_batches=100,
            limit_test_batches=100,
        )

        batch = next(iter(datamodule.init_loader))
        model(
            dense_features=batch.dense_features,
            sparse_features=batch.sparse_features,
        )
        trainer.fit(model, datamodule=datamodule)
        trainer.test(model, datamodule=datamodule)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import logging
from dataclasses import dataclass, field
from functools import partial
from typing import Any, Dict, List, Optional

import pytorch_lightning as pl
from hydra.core.config_store import ConfigStore
from omegaconf import MISSING
from pytext.contrib.pytext_lib.data.datasets import DeprecatedTsvDataset as TsvDataset
from pytext.contrib.pytext_lib.transforms import ModelTransform
from pytorch.recipes.core.conf import DataModuleConf
from pytorch.recipes.utils.config_utils import get_class_name_str
from pytorch.recipes.utils.distributed_utils import get_rank, get_world_size
from torch.utils.data import DataLoader, Dataset


logger: logging.Logger = logging.getLogger(__name__)


_DEFAULT_COLUMN_NAMES = ["text", "label"]


class DocClassificationDataModule(pl.LightningDataModule):
    def __init__(
        self,
        transform: ModelTransform,
        # Dataset args
        train_path: str,
        val_path: str,
        test_path: str,
        columns: Optional[List[str]] = None,
        column_mapping: Optional[Dict[str, str]] = None,
        delimiter: str = "\t",
        batch_size: Optional[int] = None,
        is_shuffle: bool = True,
        chunk_size: int = 1000,
        is_cycle: bool = False,
        length: Optional[int] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.train_path = train_path
        self.val_path = val_path
        self.test_path = test_path
        self.transform = transform
        # pyre-fixme[4]: Attribute must be annotated.
        self.dataset_partial = partial(
            TsvDataset,
            columns=columns or _DEFAULT_COLUMN_NAMES,
            column_mapping=column_mapping,
            delimiter=delimiter,
            batch_size=batch_size,
            is_shuffle=is_shuffle,
            transform=self.transform.transform,
            collate_fn=self.transform.collate_fn,
            chunk_size=chunk_size,
            is_cycle=is_cycle,
            length=length,
        )
        self.train_dataset: Optional[Dataset] = None
        self.val_dataset: Optional[Dataset] = None
        self.test_dataset: Optional[Dataset] = None

    def train_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        dataset = self.train_dataset
        assert dataset is not None
        return DataLoader(dataset, batch_size=None)

    def val_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        dataset = self.val_dataset
        assert dataset is not None
        return DataLoader(dataset, batch_size=None)

    def test_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        dataset = self.test_dataset
        assert dataset is not None
        return DataLoader(dataset, batch_size=None)

    def setup(self, stage: Optional[str] = None) -> None:
        world_size = get_world_size()
        rank = get_rank()
        logger.debug(f"setup for rank: {rank}, world_size: {world_size}")
        if stage == "fit" or stage is None:
            self.train_dataset = self.dataset_partial(
                path=self.train_path, rank=rank, world_size=world_size
            )
            self.val_dataset = self.dataset_partial(
                path=self.val_path, rank=rank, world_size=world_size
            )

        if stage == "test" or stage is None:
            self.test_dataset = self.dataset_partial(
                path=self.test_path, rank=rank, world_size=world_size
            )


@dataclass
class DocClassificationDataModuleConf(DataModuleConf):
    _target_: str = get_class_name_str(DocClassificationDataModule)
    train_path: str = MISSING
    val_path: str = MISSING
    test_path: str = MISSING
    columns: List[str] = field(default_factory=lambda: _DEFAULT_COLUMN_NAMES)
    column_mapping: Optional[Dict[str, str]] = None
    delimiter: str = "\t"
    batch_size: Optional[int] = None
    is_shuffle: bool = True
    chunk_size: int = 1000
    is_cycle: bool = False
    length: Optional[int] = None


cs: ConfigStore = ConfigStore.instance()

cs.store(
    group="schema/datamodule",
    name="doc_classification",
    node=DocClassificationDataModuleConf,
    package="datamodule",
)



#!/usr/bin/env python3
from itertools import combinations
from typing import List, Tuple

import torch
from torch import nn
from torchrec.fb.experimental.dynamic_embedding_modules import (
    DynamicEmbeddingBagCollection,
)
from torchrec.sparse.jagged_tensor import (
    KeyedJaggedTensor,
    KeyedTensor,
)


"""
Notations uses throughout:

F: number of sparseFeatures
D: embedding_dimension of sparse features
B: batch_size
num_features: number of dense features

"""


class SparseArch(nn.Module):
    """
    Processes the Sparse Features of SparseNN. Does Embedding Lookup for all
    EmbeddingBag and Embedding features of each collection.

    Constructor Args:
        embedding_bag_collection: DynamicEmbeddingBagCollection,

    Call Args:
        features: KeyedJaggedTensor,

    Returns:
        KeyedJaggedTensor - size F * D X B

    Example:
        >>> eb1_config = EmbeddingBagConfig(
            name="t1", embedding_dim=3, num_embeddings=10, feature_names=["f1"]
        )
        eb2_config = EmbeddingBagConfig(
            name="t2", embedding_dim=4, num_embeddings=10, feature_names=["f2"]
        )
        ebc_config = DynamicEmbeddingBagCollectionConfig(tables=[eb1_config, eb2_config])

        ebc = DynamicEmbeddingBagCollection(config=ebc_config)

        #     0       1        2  <-- batch
        # 0   [0,1] None    [2]
        # 1   [3]    [4]    [5,6,7]
        # ^
        # feature
        features = KeyedJaggedTensor.from_offsets_sync(
            keys=["f1", "f2"],
            values=torch.tensor([0, 1, 2, 3, 4, 5, 6, 7]),
            offsets=torch.tensor([0, 2, 2, 3, 4, 5, 8]),
        )

        sparse_arch(features)
    """

    def __init__(self, embedding_bag_collection: DynamicEmbeddingBagCollection) -> None:
        super().__init__()
        self.embedding_bag_collection: DynamicEmbeddingBagCollection = (
            embedding_bag_collection
        )

    def forward(
        self,
        features: KeyedJaggedTensor,
    ) -> KeyedTensor:
        return self.embedding_bag_collection(features)


class DenseArch(nn.Module):
    """
    Processes the dense features of SparseNN model. Output layer is sized to
    the embedding_dimension of the DynamicEmbeddingBagCollection embeddings

    Constructor Args:
        hidden_layer_size: int
        embedding_dim: int - the same size of the embedding_dimension of sparseArch

    Call Args:
        features: torch.Tensor  - size B X num_features

    Returns:
        torch.Tensor  - size B X D

    Example:
        >>> B = 20
        D = 3
        dense_arch = DenseArch(hidden_layer_size=10, embedding_dim=D)
        dense_embedded = dense_arch(torch.rand((B, 10)))
    """

    def __init__(self, hidden_layer_size: int, embedding_dim: int) -> None:
        super().__init__()
        self.model: nn.Module = nn.Sequential(
            nn.LazyLinear(hidden_layer_size),
            nn.ReLU(),
            nn.LazyLinear(embedding_dim),
            nn.ReLU(),
        )

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        return self.model(features)


class InteractionArch(nn.Module):
    """
    Processes the output of both SparseArch (sparse_features) and DenseArch
    (dense_features). Returns the pairwise dot product of each sparse feature pair,
    the dot product of each sparse features with the output of the dense layer,
    and the dense layer itself (all concatenated)

    Constructor Args:
        sparse_feature_names: List[str] - length of F

    Call Args:
        dense_features: torch.Tensor  - size B X D
        sparse_features: KeyedJaggedTensor - size F * D X B

    Returns:
        torch.Tensor - B X (D + F + F choose 2)

    Example:
        >>> D = 3
        B = 10
        keys = ["f1", "f2"]
        F = len(keys)
        inter_arch = InteractionArch(sparse_feature_names=keys)

        dense_features = torch.rand((B, D))

        sparse_features = KeyedTensor(
            keys=keys,
            length_per_key=[D, D],
            values=torch.rand((B, D * F)),
        )

        #  B X (D + F + F choose 2)
        concat_dense = inter_arch(dense_features, sparse_features)
    """

    def __init__(self, sparse_feature_names: List[str]) -> None:
        super().__init__()
        self.sparse_feature_names: List[str] = sparse_feature_names
        self.sparse_combinations: List[Tuple[str, str]] = list(
            combinations(sparse_feature_names, 2)
        )

    def forward(
        self, dense_features: torch.Tensor, sparse_features: KeyedTensor
    ) -> torch.Tensor:
        if len(self.sparse_feature_names) == 0:
            return dense_features

        interactions: List[torch.Tensor] = []
        # dense/sparse interaction
        # size B X F
        for feature_name in self.sparse_feature_names:
            sparse_values = sparse_features[feature_name]
            dots = torch.sum(sparse_values * dense_features, dim=1)
            # dots is size B
            interactions.append(dots)

        # sparse/sparse interaction
        # size B X (F choose 2)
        for (f1, f2) in self.sparse_combinations:
            f1_values = sparse_features[f1]
            f2_values = sparse_features[f2]
            dots = torch.sum(f1_values * f2_values, dim=1)
            interactions.append(dots)

        interactions_tensor = torch.stack(interactions).transpose(1, 0)
        return torch.cat((dense_features, interactions_tensor), dim=1)


class OverArch(nn.Module):
    """
    Final Arch of SparseNN - simple MLP over OverArch

    Constructor Args:
        output_dim: int

    Call Args:
        features: torch.Tensor

    Returns:
        torch.Tensor  - size B X output_dim

    Example:
        >>> B = 20
        D = 3
        over_arch = OverArch(output_dim=5)
        logits = over_arch(torch.rand((B, 10)))
    """

    def __init__(self, output_dim: int) -> None:
        super().__init__()
        self.model: nn.Module = nn.Sequential(
            nn.LazyLinear(output_dim),
            nn.ReLU(),
        )

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        return self.model(features)


class SimpleSparseNN(nn.Module):
    """
    Basic recsys module. Processes sparse features by
    learning pooled embeddings for each feature. Learns the relationship between
    dense features and sparse features by projecting dense features into the same
    embedding space. Also, learns the pairwise relationships between sparse features.

    The module assumes all sparse features have the same embedding dimension
    (i.e, each EmbeddingBagConfig uses the same embedding_dim)

    Constructor Args:
        embedding_bag_collection: DynamicEmbeddingBagCollection,
        hidden_layer_size: int,
        output_dim: int,

    Call Args:
        dense_features: torch.Tensor,
        sparse_features: KeyedJaggedTensor,

    Returns:
        torch.Tensor - logits with size B X output_dim

    Example:
        >>> B = 2
        D = 8

        eb1_config = EmbeddingBagConfig(
            name="t1", embedding_dim=D, num_embeddings=100, feature_names=["f1", "f3"]
        )
        eb2_config = EmbeddingBagConfig(
            name="t2",
            embedding_dim=D,
            num_embeddings=100,
            feature_names=["f2"],
        )
        ebc_config = DynamicEmbeddingBagCollectionConfig(tables=[eb1_config, eb2_config])

        ebc = DynamicEmbeddingBagCollection(config=ebc_config)
        sparse_nn = SimpleSparseNN(
            embedding_bag_collection=ebc, hidden_layer_size=20, output_dim=5
        )

        features = torch.rand((B, 100))

        #     0       1
        # 0   [1,2] [4,5]
        # 1   [4,3] [2,9]
        # ^
        # feature
        sparse_features = KeyedJaggedTensor.from_offsets_sync(
            keys=["f1", "f3"],
            values=torch.tensor([1, 2, 4, 5, 4, 3, 2, 9]),
            offsets=torch.tensor([0, 2, 4, 6, 8]),
        )

        logits = sparse_nn(
            dense_features=features,
            sparse_features=sparse_features,
        )
    """

    def __init__(
        self,
        embedding_bag_collection: DynamicEmbeddingBagCollection,
        hidden_layer_size: int,
        output_dim: int,
    ) -> None:
        super().__init__()
        assert (
            len(embedding_bag_collection.embedding_bag_configs) > 0
        ), "At least one embedding bag is required"
        for i in range(1, len(embedding_bag_collection.embedding_bag_configs)):
            conf_prev = embedding_bag_collection.embedding_bag_configs[i - 1]
            conf = embedding_bag_collection.embedding_bag_configs[i]
            assert (
                conf_prev.embedding_dim == conf.embedding_dim
            ), "All EmbeddingBagConfigs must have the same dimension"
        embedding_dim: int = embedding_bag_collection.embedding_bag_configs[
            0
        ].embedding_dim

        feature_names = []
        for conf in embedding_bag_collection.embedding_bag_configs:
            for feat in conf.feature_names:
                feature_names.append(feat)

        self.sparse_arch = SparseArch(embedding_bag_collection)
        self.dense_arch = DenseArch(
            hidden_layer_size=hidden_layer_size, embedding_dim=embedding_dim
        )
        self.inter_arch = InteractionArch(sparse_feature_names=feature_names)
        self.over_arch = OverArch(output_dim)

    def forward(
        self,
        dense_features: torch.Tensor,
        sparse_features: KeyedJaggedTensor,
    ) -> torch.Tensor:
        embedded_dense = self.dense_arch(dense_features)
        embedded_sparse = self.sparse_arch(sparse_features)
        concatenated_dense = self.inter_arch(
            dense_features=embedded_dense, sparse_features=embedded_sparse
        )
        logits = self.over_arch(concatenated_dense)
        return logits



#!/usr/bin/env python3

import argparse
import sys
from typing import List

import pytorch_lightning as pl
import torch
from aml.ml_foundation.exp_platform.e3.data.p_report_datamodule import (
    PReportTmpTableDataModule,
)
from aml.ml_foundation.exp_platform.e3.lightning.integrity_e3_classification import (
    E3BinaryClassification,
)
from stl.lightning.callbacks.model_checkpoint import ModelCheckpoint
from stl.lightning.loggers import ManifoldTensorBoardLogger
from torchrec import EmbeddingBagCollection
from torchrec.modules.embedding_configs import EmbeddingBagConfig


def parse_args(argv: List[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="torchrec + lightning app")
    parser.add_argument(
        "--epochs", type=int, default=3, help="number of epochs to train"
    )
    parser.add_argument(
        "--batch_size", type=int, default=32, help="batch size to use for training"
    )
    parser.add_argument(
        "--limit_train_batches",
        type=int,
        default=100,
        help="number of train batches",
    )
    parser.add_argument(
        "--limit_val_batches",
        type=int,
        default=100,
        help="number of val batches",
    )
    parser.add_argument(
        "--limit_test_batches",
        type=int,
        default=100,
        help="number of test batches",
    )
    parser.add_argument(
        "--manifold_bucket",
        type=str,
        help="manifold bucket for tensorboard logs and checkpointing",
        required=False,
    )
    parser.add_argument(
        "--tb_manifold_path",
        type=str,
        help="path in manifold bucket for tensorboard logs",
        required=False,
    )
    parser.add_argument("--load_path", type=str, help="checkpoint path to load from")
    parser.add_argument(
        "--checkpoint_output_path",
        type=str,
        help="path to place checkpoints",
        required=False,
    )

    return parser.parse_args(argv)


def main(argv: List[str]) -> None:
    args = parse_args(argv)

    spares_feature_keys = ["f1", "f2", "f3"]
    hash_size = 10000
    embedding_dim = 64
    ids_per_feature = 3

    datamodule = PReportTmpTableDataModule(
        use_onbox_loader=False,
        distributed_loader=False,
        mock_sparse=True,
        keys=spares_feature_keys,
        hash_size=hash_size,
        ids_per_feature=ids_per_feature,
    )

    eb_configs = [
        EmbeddingBagConfig(
            name=f"t_{feature_name}",
            embedding_dim=embedding_dim,
            num_embeddings=hash_size,
            feature_names=[feature_name],
        )
        for feature_name in spares_feature_keys
    ]

    e3_model = E3BinaryClassification(
        EmbeddingBagCollection(tables=eb_configs, device=torch.device("cpu"))
    )

    callbacks = []
    if args.checkpoint_output_path is not None:
        callbacks.append(
            ModelCheckpoint(
                save_top_k=-1,
                has_user_data=False,
                ttl_days=1,
                monitor=None,
                dirpath=args.checkpoint_output_path,
            )
        )
    if args.load_path:
        print(f"loading checkpoint: {args.load_path}...")
        e3_model.load_from_checkpoint(checkpoint_path=args.load_path)

    if args.manifold_bucket is not None and args.tb_manifold_path is not None:
        logger = ManifoldTensorBoardLogger(
            manifold_bucket=args.manifold_bucket,
            manifold_path=args.tb_manifold_path,
            has_user_data=False,
        )
    else:
        logger = None

    datamodule.setup()
    batch = next(iter(datamodule.train_dataloader())).to(e3_model.device)
    with torch.no_grad():
        e3_model(batch.dense_features, batch.sparse_features, batch.text_feature)

    datamodule._log_hyperparams = False
    e3_model._log_hyperparams = False

    trainer = pl.Trainer(
        logger=logger,
        max_epochs=args.epochs,
        callbacks=callbacks,
        limit_train_batches=args.limit_train_batches,
        limit_val_batches=args.limit_val_batches,
        limit_test_batches=args.limit_test_batches,
        weights_summary=None,
    )

    trainer.fit(e3_model, datamodule=datamodule)
    trainer.test(e3_model, datamodule=datamodule)


if __name__ == "__main__":
    main(sys.argv[1:])



def register_components() -> None:
    """
    Imports all python files in the folder to trigger the
    code to register them to Hydra's ConfigStore.
    """
    import pytorch.recipes.text.datamodule.doc_classification  # noqa



import torch


def get_rank() -> int:
    if torch.distributed.is_available() and torch.distributed.is_initialized():
        return torch.distributed.get_rank()
    return 0


def get_world_size() -> int:
    if torch.distributed.is_available() and torch.distributed.is_initialized():
        return torch.distributed.get_world_size()
    return 1


def barrier() -> None:
    """
    Wrapper over torch.distributed.barrier, returns without waiting
    if the distributed process group is not initialized instead of throwing error.
    """
    if not torch.distributed.is_available() or not torch.distributed.is_initialized():
        return
    torch.distributed.barrier()



import logging
from typing import Optional, Any, Union

import pytorch_lightning as pl
import torch
import torch.distributed as dist
from caffe2.torch.fb.distributed.utils.log_utils import getLogger
from pytorch_lightning.accelerators.accelerator import Accelerator
from pytorch_lightning.plugins import PrecisionPlugin
from pytorch_lightning.plugins.training_type.parallel import (
    ParallelPlugin,
)

logger: logging.Logger = getLogger()


class TorchrecTrainingTypePlugin(ParallelPlugin):
    """
    Lightning Trainer takes care of the operations that are related to DDP.
    However, our models are parallelization aware, which are not fully compatible to the
    given accelerators and plugins provided by Lightning.

    The torchrec plugin and acclerator bypasses the corresponding logic in Lightning.
    """

    def __init__(self) -> None:
        super().__init__()
        logger.info("Creating torchrec training type plugin")

    def broadcast(self, obj: object, src: int = 0) -> object:
        if dist.is_initialized:
            if isinstance(obj, torch.Tensor):
                dist.broadcast(obj, src)
                return obj
            else:
                object_list = [obj]
                dist.broadcast_object_list(object_list=object_list, src=src)
                return object_list[0]
        else:
            raise AssertionError(
                "Broadcast called in torchrec training type plugin w/o initializing distributed"
            )

    @property
    def root_device(self) -> torch.device:
        raise NotImplementedError()

    @property
    def should_rank_save_checkpoint(self) -> bool:
        return False

    def barrier(self, name: Optional[str] = None) -> None:
        if dist.is_initialized:
            dist.barrier()
        else:
            raise AssertionError(
                "All gather called in torchrec training type plugin w/o initializing distributed"
            )

    def all_gather(
        self,
        tensor: torch.Tensor,
        # pyre-ignore[2]: Parameter `group` has type `None` but type `Any` is specified.
        group: Optional[Any] = None,
        sync_grads: bool = False,
    ) -> torch.Tensor:
        if dist.is_initialized:
            dist.all_gather(tensor, group, sync_grads)
            return tensor
        else:
            raise AssertionError(
                "All gather called in torchrec training type plugin w/o initializing distributed"
            )

    # pyre-ignore[3]: Return type must be specified as type other than `Any`.
    def reduce(
        self,
        # pyre-ignore[2]: Parameter `tensor` must have a type other than `Any`.
        tensor: Union[Any, torch.Tensor],
        *args: Any,
        **kwargs: Any,
    ) -> Union[Any, torch.Tensor]:
        if dist.is_initialized:
            dist.all_reduce(tensor)
            return tensor
        else:
            raise AssertionError(
                "Reduce called in torchrec training type plugin w/o initializing distributed"
            )

    def model_to_device(self) -> None:
        raise NotImplementedError()

    def teardown(self) -> None:
        return None


class TorchrecAccelerator(Accelerator):
    def __init__(self) -> None:
        training_type_plugin = TorchrecTrainingTypePlugin()
        precision_plugin = PrecisionPlugin()
        super().__init__(
            precision_plugin=precision_plugin,
            training_type_plugin=training_type_plugin,
        )
        logger.info("Creating torchrec accelerator")

    def pre_dispatch(self, trainer: pl.Trainer) -> None:
        logger.info("Not doing anything in pre_dispatch of TorchrecAccelerator")
        return None

    # pyre-ignore[3]
    def batch_to_device(
        self,
        # pyre-ignore[2]
        batch: Any,
        device: Optional[torch.device] = None,
        dataloader_idx: Optional[int] = None,
    ) -> Any:
        return batch.to(self.lightning_module.device)



#!/usr/bin/env python3

from dataclasses import dataclass, field
from typing import Any, Optional, List, Dict, Union

from hydra.core.config_store import ConfigStore
from omegaconf import MISSING
from pytorch.recipes.core.conf import (
    ModuleConf,
    TrainAppConf,
    DataModuleConf,
)

cs: ConfigStore = ConfigStore.instance()


_APP_DEFAULTS: List[Union[str, Dict[str, str]]] = [
    "_self_",
    # Module
    {"schema/module": "test_module"},
    {"module/model": "linear_dummy"},
    {"module/optim": "sgd"},
    {"module/loss": "cross_entropy"},
    # Trainer
    {"schema/trainer": "trainer"},
    {"trainer": "cpu"},
    # DataModule
    {"schema/datamodule": "test_datamodule"},
    {"datamodule": "random_data"},
]


@dataclass
class TestModuleConf(ModuleConf):
    _target_: str = "pytorch.recipes.core.test_utils.test_module.TestModule"
    # pyre-fixme[4]: Attribute annotation cannot be `Any`.
    model: Any = MISSING
    # pyre-fixme[4]: Attribute annotation cannot be `Any`.
    optim: Any = MISSING
    # pyre-fixme[4]: Attribute annotation cannot be `Any`.
    loss: Any = MISSING
    # pyre-fixme[4]: Attribute annotation cannot be `Any`.
    metric: Optional[Any] = None


cs.store(
    group="schema/module", name="test_module", node=TestModuleConf, package="module"
)


@dataclass
class TestDataModuleConf(DataModuleConf):
    _target_: str = "pytorch.recipes.core.test_utils.test_module.TestDataModule"
    size: int = MISSING
    length: int = MISSING


cs.store(
    group="schema/datamodule",
    name="test_datamodule",
    node=TestDataModuleConf,
    package="datamodule",
)


@dataclass
class BaseTrainAppConf(TrainAppConf):
    _target_: str = "pytorch.recipes.core.base_train_app.BaseTrainApp"
    # pyre-fixme[4]: Attribute annotation cannot contain `Any`.
    defaults: List[Any] = field(default_factory=lambda: _APP_DEFAULTS)


cs.store(name="base_config", node=BaseTrainAppConf)



from dataclasses import dataclass
from typing import Optional, List, Iterator

import pytorch_lightning as pl
import torch

# @manual "//github/facebookresearch/hydra:hydra"
from hydra.core.config_store import ConfigStore
from pytorch.recipes.core.conf import DataModuleConf
from pytorch.recipes.rec.datamodules.commons import Batch
from pytorch.recipes.utils.config_utils import get_class_name_str
from torch.utils.data import DataLoader
from torch.utils.data.dataset import IterableDataset
from torchrec import KeyedJaggedTensor


class _RandomRecBatch:
    generator: Optional[torch.Generator]

    def __init__(
        self,
        keys: List[str],
        batch_size: int,
        hash_size: int,
        ids_per_feature: int,
        num_dense: int,
        manual_seed: Optional[int] = None,
    ) -> None:
        self.keys = keys
        self.keys_length: int = len(keys)
        self.batch_size = batch_size
        self.hash_size = hash_size
        self.ids_per_feature = ids_per_feature
        self.num_dense = num_dense

        if manual_seed is not None:
            self.generator = torch.Generator()
            # pyre-ignore[16]
            self.generator.manual_seed(manual_seed)
        else:
            self.generator = None

        self.iter_num = 0

    def __iter__(self) -> "_RandomRecBatch":
        return self

    def __next__(self) -> Batch:
        num_ids_in_batch = self.ids_per_feature * self.keys_length * self.batch_size

        sparse_features = KeyedJaggedTensor.from_offsets_sync(
            keys=self.keys,
            # pyre-ignore[28]
            values=torch.randint(
                high=self.hash_size,
                size=(num_ids_in_batch,),
                generator=self.generator,
            ),
            offsets=torch.tensor(
                list(
                    range(
                        0,
                        num_ids_in_batch + 1,
                        self.ids_per_feature,
                    )
                ),
                dtype=torch.int32,
            ),
        )

        dense_features = torch.randn(
            self.batch_size,
            self.num_dense,
            generator=self.generator,
        )
        # pyre-ignore[28]
        labels = torch.randint(
            low=0,
            high=1,
            size=(self.batch_size,),
            generator=self.generator,
        )

        batch = Batch(
            dense_features=dense_features,
            sparse_features=sparse_features,
            labels=labels,
        )
        return batch


class _RandomRecDataset(IterableDataset[Batch]):
    def __init__(
        self,
        keys: List[str],
        batch_size: int,
        hash_size: int = 100,
        ids_per_feature: int = 2,
        num_dense: int = 50,
        manual_seed: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.batch_generator = _RandomRecBatch(
            keys=keys,
            batch_size=batch_size,
            hash_size=hash_size,
            ids_per_feature=ids_per_feature,
            num_dense=num_dense,
            manual_seed=manual_seed,
        )

    def __iter__(self) -> Iterator[Batch]:
        return iter(self.batch_generator)


class RandomRecDataModule(pl.LightningDataModule):
    """
    DataModule that wraps _RandomRecDataset. This dataset generates _RandomRecBatch, or random
    batches of sparse_features in the form of KeyedJaggedTensor, dense_features and labels

    {
        "dense_features": torch.Tensor,
        "sparse_features": KeyedJaggedTensor,
        "labels": torch.Tensor,
    }
    """

    def __init__(self, manual_seed: Optional[int] = None) -> None:
        super().__init__()
        self.keys: List[str] = ["f1", "f3", "f2"]
        self.batch_size = 3
        self.manual_seed = manual_seed
        self.init_loader: DataLoader = DataLoader(
            _RandomRecDataset(
                keys=self.keys,
                batch_size=self.batch_size,
                manual_seed=self.manual_seed,
            ),
            batch_size=None,
            batch_sampler=None,
        )

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            _RandomRecDataset(
                keys=self.keys,
                batch_size=self.batch_size,
                manual_seed=self.manual_seed,
            ),
            batch_size=None,
            batch_sampler=None,
        )

    def val_dataloader(self) -> DataLoader:
        return DataLoader(
            _RandomRecDataset(
                keys=self.keys,
                batch_size=self.batch_size,
                manual_seed=self.manual_seed,
            ),
            batch_size=None,
            batch_sampler=None,
        )

    def test_dataloader(self) -> DataLoader:
        return DataLoader(
            _RandomRecDataset(
                keys=self.keys,
                batch_size=self.batch_size,
                manual_seed=self.manual_seed,
            ),
            batch_size=None,
            batch_sampler=None,
        )


@dataclass
class RandomRecDataModuleConf(DataModuleConf):
    _target_: str = get_class_name_str(RandomRecDataModule)


cs: ConfigStore = ConfigStore.instance()

cs.store(
    group="schema/datamodule",
    name="random_rec_datamodule",
    node=RandomRecDataModuleConf,
    package="datamodule",
)



#!/usr/bin/env python3

# pyre-strict

# Components to register with this config
from pytorch.recipes.text import register_components

# Register all Text DataModules and Modules so they are available for this config
register_components()



# pyre-strict

from typing import Type, TypeVar


T = TypeVar("T")

_CONFIG_ATTR = "_config_entry"


def config_entry(fn: T) -> T:
    """Decorator used to mark an object as the config entry-point for a class.

    Below is an exameple usage for this decorator. Only a single method in a
    class should be annoted as the @config_entry.

        class MyObject:
            @config_entry
            @staticmethod
            def from_config(config: MyConf) -> 'MyObject':
                ...
    """
    setattr(fn, _CONFIG_ATTR, None)
    return fn


def get_class_config_method(klass: Type[T]) -> str:
    """
    Args:
        klass: The class definition.

    Raises:
        ValueError if the klass does not define a single such method or the
        defined method is invalid.

    Returns:
        The fully qualified name of the method.
    """
    class_name = get_class_name_str(klass)
    fns = [fn for name, fn in klass.__dict__.items() if hasattr(fn, _CONFIG_ATTR)]
    if len(fns) != 1:
        raise ValueError(
            f"{class_name} has no config entrypoint. Did you use @config_entry to annotate the method?"
        )
    fn = fns[0]
    if not isinstance(fn, staticmethod):
        raise ValueError(
            f"{class_name}.{fn.__name__} is not a standalone function. Did you forget @staticmethod?"
        )
    return f"{class_name}.{fn.__func__.__name__}"


def get_class_name_str(klass: Type[T]) -> str:
    """
    Args:
        klass: The class definition.

    Returns:
        The fully qualified name of the given class.
    """
    return ".".join([klass.__module__, klass.__name__])



from . import fb_multimodal_imports  # noqa # @fb-only


def register_components() -> None:
    """
    Calls register_components() for all subfolders so we can register
    subcomponents to Hydra's ConfigStore.
    """

    from pytorch.recipes.multimodal.datamodule import (
        register_components as datamodule_components,
    )

    datamodule_components()

    from pytorch.recipes.multimodal.module import (
        register_components as module_components,
    )

    module_components()

    import pytorch.recipes.multimodal.multimodal_train_app  # noqa



#!/usr/bin/env python3

from dataclasses import dataclass
from typing import Dict, Union, Any

import pytorch_lightning as pl
import torch
import torch.nn as nn
import torchmetrics as metrics

# @manual "//github/facebookresearch/hydra:hydra"
from hydra.core.config_store import ConfigStore
from pytorch.recipes.core.conf import ModuleConf
from pytorch.recipes.rec.datamodules.commons import Batch
from pytorch.recipes.utils.config_utils import get_class_name_str
from torchrec import EmbeddingBagCollection, SimpleSparseNN
from torchrec import KeyedJaggedTensor


class SparseNNBinaryClassification(pl.LightningModule):
    def __init__(
        self,
        embedding_bag_collection: EmbeddingBagCollection,
        over_embedding_dim: int = 5,
        hidden_layer_size: int = 20,
    ) -> None:
        super().__init__()
        self.model: SimpleSparseNN = SimpleSparseNN(
            embedding_bag_collection=embedding_bag_collection,
            hidden_layer_size=hidden_layer_size,
            over_embedding_dim=over_embedding_dim,
        )
        self.loss_fn: nn.Module = nn.BCEWithLogitsLoss()
        self.accuracy: metrics.Metric = metrics.Accuracy()

    # pyre-ignore[14] - `forward` overrides method defined in `pl.core.lightning.LightningModule`
    def forward(
        self,
        dense_features: torch.Tensor,
        sparse_features: KeyedJaggedTensor,
    ) -> torch.Tensor:
        output = self.model(dense_features, sparse_features)
        return output.squeeze()

    def configure_optimizers(self) -> torch.optim.Optimizer:
        return torch.optim.Adam(self.model.parameters())

    def _step(
        self,
        batch: Batch,
        batch_idx: int,
        step_phase: str,
    ) -> torch.Tensor:
        logits = self.forward(
            dense_features=batch.dense_features,
            sparse_features=batch.sparse_features,
        )
        loss = self.loss_fn(logits, batch.labels.float())
        preds = torch.sigmoid(logits)
        accuracy = self.accuracy(preds, batch.labels)

        self.log(f"{step_phase}_accuracy", accuracy)
        self.log(f"{step_phase}_loss", loss)
        return loss

    def training_step(
        self,
        batch: Batch,
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> torch.Tensor:
        return self._step(batch, batch_idx, "train")

    def validation_step(
        self,
        batch: Batch,
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> torch.Tensor:
        return self._step(batch, batch_idx, "val")

    def test_step(
        self,
        batch: Batch,
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> torch.Tensor:
        return self._step(batch, batch_idx, "test")


@dataclass
class SparseNNBinaryClassificationModuleConf(ModuleConf):
    _target_: str = get_class_name_str(SparseNNBinaryClassification)


cs: ConfigStore = ConfigStore.instance()

cs.store(
    group="schema/module",
    name="sparsenn_binary_classification",
    node=SparseNNBinaryClassificationModuleConf,
    package="module",
)



#!/usr/bin/env python3

# pyre-strict

# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import logging
from dataclasses import dataclass
from typing import Any, Dict, Optional, Mapping

import hydra
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
import torchmetrics as metrics
from hydra.core.config_store import ConfigStore
from omegaconf import MISSING
from pytext.contrib.pytext_lib.conf import ModelConf, OptimConf, TransformConf
from pytorch.recipes.core.conf import ModuleConf
from pytorch.recipes.core.task_base import TaskBase
from pytorch.recipes.utils.config_utils import get_class_name_str
from pytorch.text.fb.transforms.model_transform import ModelTransform
from torch.nn.modules import CrossEntropyLoss

logger: logging.Logger = logging.getLogger(__name__)


class DocClassificationModule(
    TaskBase[Mapping[str, torch.Tensor], torch.Tensor, None], pl.LightningModule
):
    """
    Generic module for doc classification
    The components(model, optim etc.) can be configured and instantiated by hydra

    Note: this is a simple demo Module. Please use pytorch.recipes.text.module.doc_classification
    for training, which supports advanced features like FSDP
    """

    # See P340190896. torchscripting fails if annotated at class-level.

    def __init__(
        self,
        transform: TransformConf,
        model: ModelConf,
        optim: OptimConf,
    ) -> None:
        super().__init__()
        self.transform_conf: TransformConf = transform
        self.model_conf: ModelConf = model
        self.optim_conf: OptimConf = optim
        self.loss = CrossEntropyLoss()

        self.transform: Optional[ModelTransform] = None
        self.model: Optional[torch.nn.Module] = None
        self.optimizer: Optional[torch.optim.Optimizer] = None
        self.accuracy: Optional[metrics.Accuracy] = None
        self.fbeta: Optional[metrics.FBeta] = None

    def setup(self, stage: Optional[str]) -> None:
        """
        Called at the beginning of fit and test.
        This is a good hook when you need to build models dynamically or adjust something about them.
        This hook is called on every process when using DDP.

        Args:
            stage: either 'fit' or 'test'
        """
        # skip building model during test. Otherwise, the state dict will be re-initialized
        if stage == "fit":
            self.transform = hydra.utils.instantiate(self.transform_conf)

            transform = self.transform
            assert transform is not None

            self.model = hydra.utils.instantiate(
                self.model_conf,
                out_dim=len(transform.label_names),
                vocab=transform.vocab,
            )

            model = self.model
            assert model is not None

            self.optimizer = hydra.utils.instantiate(
                self.optim_conf,
                model.parameters(),
            )

        transform = self.transform
        assert transform is not None

        # metrics depend on transform, which is not pickle-able yet
        self.accuracy = metrics.Accuracy()
        self.fbeta = metrics.FBeta(
            num_classes=len(transform.label_names), average="macro"
        )

    # pyre-ignore[14]: This is for torchscript compatibility.
    def forward(self, batch: Dict[str, Any]) -> torch.Tensor:
        transform = self.transform
        model = self.model
        assert transform is not None and model is not None

        return model(transform(batch))

    def configure_optimizers(self) -> torch.optim.Optimizer:
        optimizer = self.optimizer
        assert optimizer is not None
        return optimizer

    def training_step(
        self,
        batch: Mapping[str, torch.Tensor],
        batch_idx: int,
        *args: Any,
        **kwargs: Any
    ) -> torch.Tensor:
        model = self.model
        assert model is not None

        logits = model(batch)
        loss = self.loss(logits, batch["label_ids"])
        self.log("train_loss", loss)
        return loss

    def validation_step(
        self,
        batch: Mapping[str, torch.Tensor],
        batch_idx: int,
        *args: Any,
        **kwargs: Any
    ) -> None:
        model = self.model
        assert model is not None
        accuracy, fbeta = self.accuracy, self.fbeta
        assert accuracy is not None and fbeta is not None

        logits = model(batch)
        loss = self.loss(logits, batch["label_ids"])
        scores = F.softmax(logits)

        accuracy(scores, batch["label_ids"])
        fbeta(scores, batch["label_ids"])
        self.log("val_loss", loss)
        self.log("val_acc", accuracy)
        self.log("val_f1", fbeta)

    def test_step(
        self,
        batch: Mapping[str, torch.Tensor],
        batch_idx: int,
        *args: Any,
        **kwargs: Any
    ) -> None:
        model = self.model
        assert model is not None
        accuracy, fbeta = self.accuracy, self.fbeta
        assert accuracy is not None and fbeta is not None

        logits = model(batch)
        loss = self.loss(logits, batch["label_ids"])
        scores = F.softmax(logits)

        accuracy(scores, batch["label_ids"])
        fbeta(scores, batch["label_ids"])
        self.log("test_loss", loss)
        self.log("test_acc", accuracy)
        self.log("test_f1", fbeta)

    def on_load_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        """
        This hook will be called before loading state_dict from a checkpoint.
        setup("fit") will built the model before loading state_dict

        Args:
            checkpoint: A dictionary with variables from the checkpoint.
        """
        self.setup("fit")


@dataclass
class DocClassificationModuleConf(ModuleConf):
    _target_: str = get_class_name_str(DocClassificationModule)
    # pyre-fixme[4]
    transform: Any = MISSING
    # pyre-fixme[4]
    model: Any = MISSING
    # pyre-fixme[4]
    optim: Any = MISSING


cs: ConfigStore = ConfigStore.instance()


cs.store(
    group="schema/module",
    name="doc_classification",
    node=DocClassificationModuleConf,
    package="module",
)



#!/usr/bin/env python3

# pyre-strict

import os
import time
import traceback
from dataclasses import dataclass
from typing import Any, Dict, Optional, TypedDict, List, Tuple

import hydra
from pytorch.recipes.core.conf import ModuleConf, DataModuleConf
from pytorch_lightning import LightningDataModule, LightningModule
from pytorch_lightning.callbacks import Callback
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.trainer import Trainer
from pytorch_lightning.utilities.types import _EVALUATE_OUTPUT, _PREDICT_OUTPUT
from stl.lightning.callbacks.model_checkpoint import ModelCheckpoint
from stl.lightning.conf.trainer import TrainerConf
from stl.lightning.loggers import ManifoldTensorBoardLogger
from stl.lightning.scuba.stl_scuba_logger import log_to_scuba, JobStatus
from stl.lightning.utilities.checkpoint import find_last_checkpoint_path
from stl.lightning.utilities.trainer_plugin_utils import get_trainer_params


@dataclass
class TrainOutput:
    tensorboard_log_dir: Optional[str] = None


class TestOutput(TypedDict):
    pass


class BaseTrainApp:
    """
    A training recipe that contains all necessary parts to train a model.
    One can easily start a trainig flow with this training application.
    To use the interface, create your own TrainApp and subclass from the BaseTrainApp.
    You also need to subclass YourTaskConfig from ModuleConf.
    """

    module_conf: ModuleConf
    module: LightningModule
    datamodule_conf: Optional[DataModuleConf]
    datamodule: Optional[LightningDataModule]
    trainer_conf: TrainerConf
    log_dir: Optional[str]
    root_dir: Optional[str]

    def __init__(
        self,
        module: ModuleConf,
        trainer: TrainerConf,
        datamodule: Optional[DataModuleConf] = None,
    ) -> None:
        super().__init__()
        self.module_conf = module
        self.module = self.get_lightning_module()

        self.datamodule_conf = datamodule
        self.datamodule = self.get_data_module()

        self.trainer_conf = trainer
        self.log_dir = None
        self.root_dir = None

    def get_lightning_module(self) -> LightningModule:
        """
        Override this method to instantiate a LightningModule
        """
        return hydra.utils.instantiate(self.module_conf, _recursive_=False)

    def get_data_module(self) -> Optional[LightningDataModule]:
        """
        Override this method to instantiate a LightningDataModule
        """
        if self.datamodule_conf:
            return hydra.utils.instantiate(self.datamodule_conf, _recursive_=False)

        return None

    def get_callbacks(self) -> List[Callback]:
        """
        Override this method to return a list of callbacks to be passed into Trainer
        You can add additional ModelCheckpoint here
        """
        return []

    def get_logger(self) -> TensorBoardLogger:
        """
        Override this method to return a logger for trainer
        TODO: T88650989 set different default logger for OSS and FB TrainApp
        """
        return ManifoldTensorBoardLogger()

    def get_default_model_checkpoint(self) -> ModelCheckpoint:
        """
        Override this method to return a default ModelCheckpoint callback.
        Note: If you want to use more than 1 ModelCheckpoint callback, add it through
        get_callbacks() function.
        """
        dirpath: Optional[str] = None
        root_dir = self.root_dir
        if root_dir:
            dirpath = os.path.join(root_dir, ModelCheckpoint.CHECKPOINT_PATH_SUFFIX)

        return ModelCheckpoint(
            # will auto generate dirpath if not provided
            dirpath=dirpath,
            save_top_k=-1,
            has_user_data=False,
            ttl_days=1,
            monitor=None,
        )

    def _get_trainer(self) -> Tuple[Trainer, Dict[str, Any]]:
        trainer_params = self._init_trainer_params()
        self._set_trainer_params(trainer_params)

        # log trainer params
        log_params = dict(trainer_params)
        log_params["oncall_team"] = "pt_lightning"
        log_params["run_status"] = JobStatus.RUNNING.value
        log_to_scuba(**log_params)

        return Trainer(**trainer_params), log_params

    def _init_trainer_params(self) -> Dict[str, Any]:
        return get_trainer_params(self.trainer_conf)

    def _set_trainer_params(
        self,
        trainer_params: Dict[str, Any],
    ) -> None:
        # set default logger if not specified
        # if logger=False, do not add a logger
        if trainer_params.get("logger", True):
            logger = self.get_logger()
            trainer_params["logger"] = logger
            self.log_dir = logger.log_dir
            self.root_dir = logger.root_dir

        callbacks = trainer_params.get("callbacks", [])
        callbacks.extend(self.get_callbacks())

        # create default model checkpoint callback unless disabled
        if trainer_params.get("checkpoint_callback", True):
            checkpoint_callback = self.get_default_model_checkpoint()
            callbacks.append(checkpoint_callback)

            # auto-resume from last default checkpoint
            ckpt_path = checkpoint_callback.dirpath
            if not trainer_params.get("resume_from_checkpoint") and ckpt_path:
                last_checkpoint = find_last_checkpoint_path(ckpt_path)
                trainer_params["resume_from_checkpoint"] = last_checkpoint

        trainer_params["callbacks"] = callbacks

    def train(self) -> TrainOutput:
        trainer, log_params = self._get_trainer()

        start_time = time.monotonic()
        got_exception = None
        try:
            trainer.fit(self.module, datamodule=self.datamodule)
        except Exception as ex:
            got_exception = ex

        # log trainer status to Scuba and Hive
        total_run_time = int(time.monotonic() - start_time)
        log_params["global_rank"] = trainer.global_rank
        log_params["world_size"] = trainer.world_size
        log_params["total_run_time"] = total_run_time
        if got_exception is None:
            log_params["run_status"] = JobStatus.COMPLETED.value
            log_to_scuba(**log_params)
        else:
            log_params["error_message"] = str(got_exception)
            log_params["stacktrace"] = traceback.format_stack()
            log_params["run_status"] = JobStatus.FAILED.value
            log_to_scuba(**log_params)
            raise got_exception

        return TrainOutput(tensorboard_log_dir=self.log_dir)

    def test(self) -> _EVALUATE_OUTPUT:
        trainer, _ = self._get_trainer()
        return trainer.test(self.module, datamodule=self.datamodule)

    def predict(self) -> Optional[_PREDICT_OUTPUT]:
        trainer, _ = self._get_trainer()
        return trainer.predict(self.module, datamodule=self.datamodule)



from dataclasses import dataclass, field
from typing import List, Any, Dict, Optional

from hydra.core.config_store import ConfigStore
from pytorch.recipes.core.base_train_app import BaseTrainApp
from pytorch.recipes.core.conf import ModuleConf, DataModuleConf
from pytorch.recipes.core.conf.base_config import BaseTrainAppConf
from pytorch.recipes.rec.datamodules.criteo_datamodule import CriteoDataModule
from pytorch.recipes.rec.modules.sparsenn_classification import (
    SparseNNBinaryClassification,
    SparseNNBinaryClassificationModuleConf,
)
from pytorch.recipes.utils.config_utils import get_class_name_str
from pytorch_lightning import LightningDataModule, LightningModule
from stl.lightning.conf.trainer import TrainerConf
from torchrec import EmbeddingBagCollection
from torchrec.datasets.criteo import DEFAULT_CAT_NAMES
from torchrec.modules.embedding_configs import EmbeddingBagConfig


defaults: List[Dict[str, str]] = [
    {"trainer": "rec_cpu"},
]


class SparseNNBinaryClassificationTrainApp(BaseTrainApp):
    """
    Train App for torchrec classification
    """

    datamodule: CriteoDataModule

    def __init__(
        self,
        module: ModuleConf,
        trainer: TrainerConf,
        datamodule: Optional[DataModuleConf] = None,
        num_embeddings: int = 100000,
    ) -> None:
        self.num_embeddings = num_embeddings
        super().__init__(module, trainer, datamodule)

    def get_lightning_module(self) -> LightningModule:
        keys = DEFAULT_CAT_NAMES
        embedding_dim = 64

        eb_configs = [
            EmbeddingBagConfig(
                name=f"t_{feature_name}",
                embedding_dim=embedding_dim,
                num_embeddings=self.num_embeddings,
                feature_names=[feature_name],
            )
            for feature_name in keys
        ]
        return SparseNNBinaryClassification(EmbeddingBagCollection(tables=eb_configs))

    def get_data_module(self) -> LightningDataModule:
        return CriteoDataModule(
            num_days=10,
            num_days_test=3,
            num_embeddings=self.num_embeddings,
        )

    # TODO: find a better way to deal with lazy modules in lightning w/o disabling weights_summary


@dataclass
class SparseNNBinaryClassificationTrainAppConf(BaseTrainAppConf):
    _target_: str = get_class_name_str(SparseNNBinaryClassificationTrainApp)
    # pyre-ignore [4]: Attribute `defaults` of class `SparseNNBinaryClassificationTrainAppConf` is used as type `List[Dict[str, str]]` and must have a type that does not contain `Any`.
    defaults: List[Any] = field(default_factory=lambda: defaults)
    trainer: TrainerConf = TrainerConf()
    # Dummy conf. overridden in get_lightning_module()
    module: SparseNNBinaryClassificationModuleConf = (
        SparseNNBinaryClassificationModuleConf()
    )


cs: ConfigStore = ConfigStore.instance()

cs.store(
    name="rec_classification_train_app_conf",
    node=SparseNNBinaryClassificationTrainAppConf,
)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved


# pyre-ignore-all-errors

# Ignore all pyre errors in this file for now since it should be refactored to use Standard DocClassificationModule soon
import logging
from typing import Any, Dict, cast

import hydra
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
import torchmetrics as metrics
from fairscale.nn import wrap
from pytext.contrib.pytext_lib.conf import (
    DocClassificationDataModuleConf,
    DocTransformConf,
    ModelConf,
    OptimConf,
    SchedulerConf,
    TransformConf,
)
from pytext.contrib.pytext_lib.utils.usage import log_class_usage
from pytext.utils.file_io import PathManager
from pytorch.recipes.text.module.common import CheckpointMixin
from torch.serialization import default_restore_location

logger: logging.Logger = logging.getLogger(__name__)


class DocClassificationTask(CheckpointMixin, pl.LightningModule):
    """
    Generic task for doc classification
    The components(model, optim etc.) can be configured and instantiated by hydra

    Example:
            .. code-block:: python

                with initialize_config_module("pytext.contrib.pytext_lib.conf"):
                    cfg = compose(
                        config_name="pytext_config",
                        overrides=[
                            "task/transform=doc_transform",
                            "task/datamodule=doc_classification_dummy",
                            "task/model=xlmr_dummy",
                            "task/optim=adamw",
                            "trainer=cpu",
                            "trainer.max_epochs=10",
                        ],
                    )
                task = hydra.utils.instantiate(cfg.task, _recursive_=False)
                transform = hydra.utils.instantiate(cfg.task.transform, _recursive_=False)
                datamodule = hydra.utils.instantiate(cfg.task.datamodule, transform=transform, _recursive_=False)
                trainer = Trainer(**cfg.trainer)
                trainer.fit(task, datamodule=datamodule)
                trainer.test(task, datamodule=datamodule)
    """

    def __init__(
        self,
        transform: TransformConf,
        datamodule: DocClassificationDataModuleConf,
        model: ModelConf,
        optim: OptimConf,
        scheduler: SchedulerConf = None,
        enable_auc_metric: bool = False,
        enable_configure_sharded_model: bool = False,
        encoder_eval_mode: bool = False,
        pretrained_checkpoint_path: str = "",
    ) -> None:
        super().__init__()
        self.transform_conf = transform
        # Keeping datamodule_conf for now so we don't refactor TaskConf
        self.datamodule_conf = datamodule
        self.model_conf = model
        self.optim_conf = optim
        self.scheduler_conf = scheduler
        self.loss = torch.nn.CrossEntropyLoss()
        self.enable_configure_sharded_model = enable_configure_sharded_model
        self.enable_auc_metric = enable_auc_metric
        self.encoder_eval_mode = encoder_eval_mode
        self.pretrained_checkpoint_path = pretrained_checkpoint_path
        log_class_usage(self.__class__)

        # metrics depend on transform, which is not pickle-able yet
        num_classes = len(hydra.utils.instantiate(self.transform_conf).label_names)
        self.valid_acc = metrics.Accuracy()
        self.valid_f1 = metrics.FBeta(num_classes=num_classes, average="macro")
        self.test_acc = metrics.Accuracy()
        self.test_f1 = metrics.FBeta(num_classes=num_classes, average="macro")
        if self.enable_auc_metric:
            self.valid_aucroc = metrics.AUROC(
                num_classes=num_classes, compute_on_step=False
            )
            self.test_aucroc = metrics.AUROC(
                num_classes=num_classes, compute_on_step=False
            )

    def setup(self, stage: str) -> None:
        """
        Called at the beginning of fit and test.
        This is a good hook when you need to build models dynamically or adjust something about them.
        This hook is called on every process when using DDP.

        Args:
            stage: either 'fit' or 'test'
        """
        # skip building model during test. Otherwise, the state dict will be re-initialized
        if stage == "test":
            return
        # resetting call_configure_sharded_model_hook attribute so that we could configure model
        self.call_configure_sharded_model_hook = False

        self.transform = hydra.utils.instantiate(self.transform_conf)
        num_classes = len(self.transform.label_names)
        self.model = hydra.utils.instantiate(
            self.model_conf,
            out_dim=num_classes,
            vocab=self.transform.vocab,
        )

        if self.pretrained_checkpoint_path:
            checkpoint_dict = torch.load(
                PathManager.get_local_path(self.pretrained_checkpoint_path),
                map_location=lambda s, l: default_restore_location(s, "cpu"),
            )
            self.load_state_dict(checkpoint_dict["state_dict"])
            print(f"Loaded state dict from {self.pretrained_checkpoint_path}")

    def forward(self, batch: Dict[str, Any]) -> torch.Tensor:
        return self.model(self.transform(batch))

    def training_step(self, batch, batch_idx):
        logits = self.model(batch)
        loss = self.loss(logits, batch["label_ids"])
        self.log("train_loss", loss)
        self.log("learning_rate", self.optimizer.param_groups[0]["lr"])
        return loss

    def validation_step(self, batch, batch_idx) -> None:
        logits = self.model(batch)
        assert "label_ids" in batch, batch
        loss = self.loss(logits, batch["label_ids"])
        scores = F.softmax(logits)
        # We should call update here as we are not using the output
        self.valid_acc.update(scores, batch["label_ids"])
        self.valid_f1.update(scores, batch["label_ids"])
        self.log("val_loss", loss)
        self.log("val_acc", self.valid_acc)
        self.log("val_f1", self.valid_f1)
        if self.enable_auc_metric:
            self.valid_aucroc.update(scores, batch["label_ids"])
            self.log("val_aucroc", self.valid_aucroc, metric_attribute="valid_aucroc")

    def test_step(self, batch, batch_idx) -> None:
        logits = self.model(batch)
        loss = self.loss(logits, batch["label_ids"])
        scores = F.softmax(logits)
        # We should call update here as we are not using the output
        self.test_acc.update(scores, batch["label_ids"])
        self.test_f1.update(scores, batch["label_ids"])
        self.log("test_loss", loss)
        self.log("test_acc", self.test_acc)
        self.log("test_f1", self.test_f1)
        if self.enable_auc_metric:
            self.test_aucroc.update(scores, batch["label_ids"])
            self.log("test_aucroc", self.test_aucroc, metric_attribute="test_aucroc")

    def configure_optimizers(
        self,
    ) -> Any:
        # we put optimizer instantiation in `configure_optimizers` as we
        # need wrapped model's parameters
        self.optimizer = hydra.utils.instantiate(
            self.optim_conf, self.model.parameters()
        )
        self.scheduler = hydra.utils.instantiate(self.scheduler_conf, self.optimizer)
        if self.scheduler is not None:
            return [self.optimizer], [{"scheduler": self.scheduler, "interval": "step"}]
        else:
            return self.optimizer

    def configure_sharded_model(self) -> None:
        if not self.enable_configure_sharded_model:
            return
        # wrap is annotation, which will be enabled inside
        # `enable_wrap` context manager
        if hasattr(self.model, "configure_sharded_model"):
            self.model.configure_sharded_model()
        self.model = wrap(self.model)

        num_model_parameters = sum(p.numel() for p in self.model.parameters())
        logger.info(
            f"After configuring sharded model, model's number of parameters is: {num_model_parameters}."
        )

    def on_load_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        """
        This hook will be called before loading state_dict from a checkpoint.
        setup("fit") will built the model before loading state_dict

        Args:
            checkpoint: A dictionary with variables from the checkpoint.
        """
        self.setup("fit")

    def on_train_epoch_start(self) -> None:
        if self.encoder_eval_mode and hasattr(self.model, "encoder"):
            self.model.encoder.eval()
            logger.info("Set encoder to eval mode.")



#!/usr/bin/env python3
import os
from typing import Any, List, TypeVar, cast

import pytorch_lightning as pl
import torch
import torch.distributed as dist
import torch.nn as nn
import torchmetrics as metrics
from pytorch.recipes.rec.datamodules.criteo_datamodule import Batch
from torchrec import EmbeddingBagCollection, SimpleSparseNN
from torchrec import KeyedJaggedTensor
from torchrec.distributed.embedding import EmbeddingBagCollectionSharder
from torchrec.distributed.embedding_types import EmbeddingComputeKernel
from torchrec.distributed.model_parallel import DistributedModelParallel
from torchrec.distributed.types import ShardingType, ModuleSharder
from torchrec.optim.keyed import KeyedOptimizerWrapper

M = TypeVar("M", bound=nn.Module)


class TWEmbeddingBagCollectionSharder(EmbeddingBagCollectionSharder[M]):
    @property
    def sharding_types(self) -> List[str]:
        return [ShardingType.TABLE_WISE.value, ShardingType.DATA_PARALLEL.value]

    def compute_kernels(self, sharding_type: str, device: torch.device) -> List[str]:
        return [
            # restrict to DENSE until T98586878 is fixed
            EmbeddingComputeKernel.DENSE.value,
        ]


class ShardedSparseNNBinaryClassification(pl.LightningModule):
    def __init__(
        self,
        embedding_bag_collection: EmbeddingBagCollection,
        over_embedding_dim: int = 5,
        hidden_layer_size: int = 20,
    ) -> None:
        super().__init__()

        rank = int(os.environ["LOCAL_RANK"])
        if torch.cuda.is_available():
            device = torch.device(f"cuda:{rank}")
            backend = "nccl"
            torch.cuda.set_device(device)
        else:
            device = torch.device("cpu")
            backend = "gloo"

        if not torch.distributed.is_initialized():
            dist.init_process_group(backend=backend)
        self.to(device=device)

        model = SimpleSparseNN(
            embedding_bag_collection=embedding_bag_collection,
            hidden_layer_size=hidden_layer_size,
            over_embedding_dim=over_embedding_dim,
            dense_device=device,
            sparse_device=torch.device("meta"),
        )
        sharders = cast(
            List[ModuleSharder[nn.Module]],
            [TWEmbeddingBagCollectionSharder()],
        )

        self.model = DistributedModelParallel(
            module=model,
            device=device,
            sharders=sharders,
            init_data_parallel=False,
        )
        self.loss_fn: nn.Module = nn.BCEWithLogitsLoss()
        self.accuracy: metrics.Metric = metrics.Accuracy().to(device=device)

    # pyre-ignore[14] - `forward` overrides method defined in `pl.core.lightning.LightningModule`
    def forward(
        self,
        dense_features: torch.Tensor,
        sparse_features: KeyedJaggedTensor,
    ) -> torch.Tensor:
        logits = self.model(dense_features, sparse_features)
        return logits.squeeze()

    def configure_optimizers(self) -> torch.optim.Optimizer:
        return KeyedOptimizerWrapper(
            dict(self.model.named_parameters()),
            lambda params: torch.optim.SGD(params, lr=0.01),
        )

    def _step(
        self,
        batch: Batch,
        batch_idx: int,
        step_phase: str,
    ) -> torch.Tensor:
        logits = self.forward(
            dense_features=batch.dense_features,
            sparse_features=batch.sparse_features,
        )
        loss = self.loss_fn(logits, batch.labels.float())
        preds = torch.sigmoid(logits)
        accuracy = self.accuracy(preds, batch.labels)

        self.log(f"{step_phase}_accuracy", accuracy)
        self.log(f"{step_phase}_loss", loss)
        return loss

    def training_step(
        self,
        batch: Batch,
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> torch.Tensor:
        return self._step(batch, batch_idx, "train")

    def validation_step(
        self,
        batch: Batch,
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> torch.Tensor:
        return self._step(batch, batch_idx, "val")

    def test_step(
        self,
        batch: Batch,
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> torch.Tensor:
        return self._step(batch, batch_idx, "test")



#!/usr/bin/env python3
import os

import pytorch_lightning as pl
import torch
import torch.distributed as dist


class IGReelsClipsTab(pl.LightningModule):
    def __init__(
        self,
    ) -> None:
        super().__init__()

        rank = int(os.environ["LOCAL_RANK"])
        if torch.cuda.is_available():
            device = torch.device(f"cuda:{rank}")
            backend = "nccl"
            torch.cuda.set_device(device)
        else:
            device = torch.device("cpu")
            backend = "gloo"

        dist.init_process_group(backend=backend)
        self.to(device=device)



#!/usr/bin/env python3
from functools import lru_cache
from typing import Dict, Tuple, List, Mapping

import hydra
import torch
from omegaconf import OmegaConf
from pytorch.recipes.multimodal.module import EmbeddingConf, EncoderConf, FusionConf
from pytorch.recipes.utils.mixup_utils import MixupScheme


class DictEncoderWrap(torch.nn.Module):
    """
    Wraps an encoder to make it input and output tensor dicts.

    The wrapped encoder has single-tensor input and return type.
    """

    def __init__(self, module: torch.nn.Module, channel_name: str) -> None:
        super().__init__()
        self.module = module
        self.channel_name = channel_name
        # pyre-fixme[4]: Missing attribute annotation [4]: Attribute `output_dim` of class
        # `DictEncoderWrap` has type `Union[torch.Tensor, torch.nn.Module]` but no type is specified.
        self.output_dim = getattr(self.module, "output_dim", None)

    def forward(self, in_batch: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        input = in_batch[self.channel_name]
        output = self.module(input)
        return {self.channel_name: output}


def validate_multimodal_config(
    embeddings_conf: List[EmbeddingConf],
    encoders_conf: List[EncoderConf],
    fusion_conf: List[FusionConf],
) -> None:
    """
    Validate input configurations for embeddings, encoders and fusions
    """

    # check encoder channel names are unique
    encoders_channel_names = set()
    for encoder in encoders_conf:
        if encoder.channel_name in encoders_channel_names:
            raise AssertionError(
                f"Duplicate Channel name {encoder.channel_name} in encoders list"
            )
        encoders_channel_names.add(encoder.channel_name)

    # check whether an encoder exists for each embedding
    for embedding in embeddings_conf:
        if embedding.channel_name not in encoders_channel_names:
            raise AssertionError(
                f"Embedding {embedding.channel_name} has no corresponding encoder"
            )

    # check dag of components is correctly configured
    fusion_names = set()
    for fusion in fusion_conf:
        for input_name in fusion.input_encoder_names:
            if (
                input_name not in encoders_channel_names
                and input_name not in fusion_names
            ):
                raise AssertionError(
                    f"Input name {input_name} not found in encoders list. "
                    "If you are using a fusion as input to another fusion, please make "
                    "sure the order is correct"
                )
        fusion_names.add(fusion.fusion_name)


def build_encoders(
    encoders_conf: List[EncoderConf],
    embeddings_conf: List[EmbeddingConf],
) -> Dict[str, torch.nn.Module]:
    """
    Builds encoders given conf objects for encoders and corresponding
    embeddings (if applicable).

    Returns a module dict of encoders keyed on the channel name
    """
    embeddings = {}
    if embeddings_conf:
        embeddings = {
            embedding.channel_name: hydra.utils.instantiate(embedding.embedding_args)
            for embedding in embeddings_conf
        }
    encoders = torch.nn.ModuleDict()
    for encoder in encoders_conf:
        if encoder.channel_name in embeddings:
            encoders[encoder.channel_name] = hydra.utils.instantiate(
                encoder.encoder_args, embedding=embeddings[encoder.channel_name]
            )

        else:
            encoders[encoder.channel_name] = hydra.utils.instantiate(
                encoder.encoder_args
            )

        if not encoder.dict_input:
            encoders[encoder.channel_name] = DictEncoderWrap(
                encoders[encoder.channel_name], encoder.channel_name
            )

    return encoders


def build_fusion(
    fusion_conf: List[FusionConf],
    encoders: Mapping[str, torch.nn.Module],
) -> Tuple[Dict[str, torch.nn.Module], Dict[str, List[str]]]:
    """
    Builds fusion modules given conf objects for fusion and
    module dict of encoders.

    Returns a module dict of fusion modules keyed on unique fusion name
    and a dict mapping fusion to input encoder names.
    """

    fusions = torch.nn.ModuleDict()
    fusion_inputs = {}
    for fusion in fusion_conf:
        fusion_inputs[fusion.fusion_name] = OmegaConf.to_container(
            fusion.input_encoder_names
        )
        if "encoders_dim_dict" in fusion.fusion_args:
            fusions[fusion.fusion_name] = hydra.utils.instantiate(fusion.fusion_args)
        else:
            encoders_dim_dict = {}
            for encoder_name in fusion.input_encoder_names:
                encoders_dim_dict[encoder_name] = encoders[encoder_name].output_dim
            fusions[fusion.fusion_name] = hydra.utils.instantiate(
                fusion.fusion_args, encoders_dim_dict=encoders_dim_dict
            )
    return fusions, fusion_inputs


class MultimodalClassificationMixupScheme(MixupScheme):
    INPUT = "INPUT"
    AFTERENCODING = "AFTERENCODING"
    AFTERFUSION = "AFTERFUSION"
    AFTERCLASSIFIER = "AFTERCLASSIFIER"
    RANDOM = "RANDOM"

    @classmethod
    @lru_cache
    def get_all_except_RANDOM(cls) -> List[MixupScheme]:
        return [scheme for scheme in list(cls) if scheme is not cls.RANDOM]



#!/usr/bin/env python3

from typing import (
    Any,
    Iterable,
    List,
    Mapping,
    Optional,
    Union,
    cast,
)

import pytorch_lightning as pl
import torch
from pytorch.recipes.rec.datamodules.commons import Batch
from pytorch.recipes.rec.datamodules.samplers.undersampler import ProportionUnderSampler
from torch.utils.data import DataLoader, IterDataPipe
from torchrec.datasets.criteo import (
    DEFAULT_CAT_NAMES,
    DEFAULT_INT_NAMES,
    DEFAULT_LABEL_NAME,
)
from torchrec.datasets.criteo import criteo_terabyte
from torchrec.datasets.utils import rand_split_train_val
from torchrec.fb.datasets.utils import enable_manifold_read
from torchrec.sparse.jagged_tensor import KeyedJaggedTensor


def _transform(
    batch: Mapping[str, Union[Iterable[str], torch.Tensor]],
    num_embeddings: int = 100000,
) -> Batch:
    cat_list: List[torch.Tensor] = []
    for col_name in DEFAULT_INT_NAMES:
        val = cast(torch.Tensor, batch[col_name])
        # mininum value in criteo dataset of int features is -1 so we add 2 before taking log
        cat_list.append((torch.log(val + 2)).unsqueeze(0).T)
    dense_features = torch.cat(
        cat_list,
        dim=1,
    )

    kjt_values: List[int] = []
    kjt_lengths: List[int] = []
    for col_name in DEFAULT_CAT_NAMES:
        values = cast(Iterable[str], batch[col_name])
        for value in values:
            if value:
                kjt_values.append(int(value, 16) % num_embeddings)
                kjt_lengths.append(1)
            else:
                kjt_lengths.append(0)

    sparse_features = KeyedJaggedTensor.from_lengths_sync(
        DEFAULT_CAT_NAMES,
        torch.tensor(kjt_values),
        torch.tensor(kjt_lengths, dtype=torch.int32),
    )
    labels = batch[DEFAULT_LABEL_NAME]
    assert isinstance(labels, torch.Tensor)

    return Batch(
        dense_features=dense_features,
        sparse_features=sparse_features,
        labels=labels,
    )


# Setup function to enable manifold reading for each DataLoader worker process.
def _enable_manifold_read(_worker_id: int) -> None:
    enable_manifold_read()


class CriteoDataModule(pl.LightningDataModule):
    """`DataModule for Criteo 1TB Click Logs <https://ailab.criteo.com/download-criteo-1tb-click-logs-dataset/>`_ Dataset
    Args:
        num_days: number of days (out of 25) of data to use for train/validation
        num_days_test: number of days (out of 25) of data to use for testing
        num_embeddings: the number of embeddings (hash size) of the categorical (sparse) features
        batch_size: int
        num_workers: int
        train_percent: percent of data to use for training vs validation- 0.0 - 1.0
        read_chunk_size: int
        dataset_path: Path to the criteo dataset. Defaults to torchrec maintained manifold path
        undersampling_rate: 0.0 - 1.0. Desired proportion of zero-labeled samples to
            retain (i.e. undersampling zero-labeled rows). Ex. 0.3 indicates only 30%
            of the rows with label 0 will be kept. All rows with label 1 will be kept.
            Default: None, indicating no undersampling.
        seed: Random seed for reproducibility. Default: None.

    Examples:
        >>> dm = CriteoDataModule(num_days=1, batch_size=3, num_days_test=1)
        >>> dm.setup()
        >>> train_batch = next(iter(dm.train_dataloader()))
    """

    def __init__(
        self,
        num_days: int = 1,
        num_days_test: int = 0,
        num_embeddings: int = 100000,
        batch_size: int = 32,
        train_percent: float = 0.8,
        num_workers: int = 0,
        read_chunk_size: int = 100000,
        dataset_path: str = "manifold://torchrec/tree/datasets/criteo",
        undersampling_rate: Optional[float] = None,
        seed: Optional[int] = None,
    ) -> None:
        super().__init__()
        if not (1 <= num_days <= 24):
            raise ValueError(
                f"Dataset has only 24 days of data. User asked for {num_days} days"
            )
        if not (0 <= num_days_test <= 24):
            raise ValueError(
                f"Dataset has only 24 days of data. User asked for {num_days_test} days"
            )
        if not (num_days + num_days_test <= 24):
            raise ValueError(
                f"Dataset has only 24 days of data. User asked for {num_days} train days and {num_days_test} test days"
            )
        if not (0.0 <= train_percent <= 1.0):
            raise ValueError(f"train_percent {train_percent} must be between 0 and 1")

        # TODO handle more workers for IterableDataset

        self.batch_size = batch_size
        self._num_workers = num_workers
        self._read_chunk_size = read_chunk_size
        self._num_days = num_days
        self._num_days_test = num_days_test
        self.num_embeddings = num_embeddings
        self._train_percent = train_percent
        self._dataset_path = dataset_path
        self._undersampling_rate = undersampling_rate
        self._seed = seed

        self._train_datapipe: Optional[IterDataPipe] = None
        self._val_datapipe: Optional[IterDataPipe] = None
        self._test_datapipe: Optional[IterDataPipe] = None
        self.keys: List[str] = DEFAULT_CAT_NAMES

    def _create_datapipe(self, day_range: Iterable[int]) -> IterDataPipe:
        paths = [f"{self._dataset_path}/day_{day}.tsv" for day in day_range]
        datapipe = criteo_terabyte(
            paths,
            # this is important because without it, the reader will attempt to synchronously download the whole file...
            read_chunk_size=self._read_chunk_size,
        )

        undersampling_rate = self._undersampling_rate
        if undersampling_rate is not None:
            datapipe = ProportionUnderSampler(
                datapipe,
                CriteoDataModule._get_label,
                {0: undersampling_rate, 1: 1.0},
                seed=self._seed,
            )

        return datapipe

    @staticmethod
    # pyre-ignore[2, 3]
    def _get_label(row: Any) -> Any:
        return row["label"]

    def _batch_collate_transform(self, datapipe: IterDataPipe) -> IterDataPipe:
        return (
            datapipe.batch(self.batch_size)
            .collate()
            .map(_transform, fn_kwargs={"num_embeddings": self.num_embeddings})
        )

    def setup(self, stage: Optional[str] = None) -> None:
        # TODO: move enable_manifold_read() call in fb only when move to OSS (potentially subclass)
        enable_manifold_read()
        if stage == "fit" or stage is None:
            datapipe = self._create_datapipe(range(self._num_days))
            train_datapipe, val_datapipe = rand_split_train_val(
                datapipe, self._train_percent
            )
            self._train_datapipe = self._batch_collate_transform(train_datapipe)
            self._val_datapipe = self._batch_collate_transform(val_datapipe)

        if stage == "test" or stage is None:
            datapipe = self._create_datapipe(
                range(self._num_days, self._num_days + self._num_days_test)
            )
            self._test_datapipe = self._batch_collate_transform(datapipe)

    def _create_dataloader(self, datapipe: IterDataPipe) -> DataLoader:
        return DataLoader(
            datapipe,
            num_workers=self._num_workers,
            batch_size=None,
            batch_sampler=None,
            # TODO: move enable_manifold_read() call in fb only when move to OSS (potentially subclass)
            worker_init_fn=_enable_manifold_read,
        )

    def train_dataloader(self) -> DataLoader:
        datapipe = self._train_datapipe
        assert isinstance(datapipe, IterDataPipe)
        return self._create_dataloader(datapipe)

    def val_dataloader(self) -> DataLoader:
        datapipe = self._val_datapipe
        assert isinstance(datapipe, IterDataPipe)
        return self._create_dataloader(datapipe)

    def test_dataloader(self) -> DataLoader:
        datapipe = self._test_datapipe
        assert isinstance(datapipe, IterDataPipe)
        return self._create_dataloader(datapipe)



import unittest
import uuid

from pytorch.recipes.rec.sparsenn_binary_classification_main import main
from torch.distributed.launcher.api import elastic_launch, LaunchConfig
from torchrec.tests import utils


class MainTest(unittest.TestCase):
    @classmethod
    def _run_trainer(cls) -> None:
        main(
            [
                "--limit_train_batches",
                "5",
                "--limit_val_batches",
                "5",
                "--limit_test_batches",
                "5",
            ]
        )

    @utils.skip_if_asan
    def test_main_function(self) -> None:
        lc = LaunchConfig(
            min_nodes=1,
            max_nodes=1,
            nproc_per_node=2,
            run_id=str(uuid.uuid4()),
            # TODO T100608035 replace with c10d when fixed
            rdzv_backend="zeus",
            rdzv_endpoint="",
            start_method="spawn",
            monitor_interval=1,
            max_restarts=0,
        )

        elastic_launch(config=lc, entrypoint=self._run_trainer)()



#!/usr/bin/env python3

# pyre-strict


def register_components() -> None:
    """
    Calls register_components() for all subfolders so we can register
    subcomponents to Hydra's ConfigStore.
    """

    from pytorch.recipes.text.datamodule import (
        register_components as datamodule_components,
    )

    datamodule_components()

    from pytorch.recipes.text.module import register_components as module_components

    module_components()

    import pytorch.recipes.text.doc_classification_train_app  # noqa



# pyre-strict

from abc import ABC, abstractmethod
from typing import Any, Generic, Iterable, TypeVar, Tuple, Union

from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler

TBatch = TypeVar("TBatch")
TTrainReturn = TypeVar("TTrainReturn")
TTestReturn = TypeVar("TTestReturn")


class TaskBase(ABC, Generic[TBatch, TTrainReturn, TTestReturn]):
    """Abstract base class for a Task.

    This allows us to be more opinionated than the base LightningModule about
    what methods must be implemented in a Standard Tasks. Furthermore, if
    these methods are not implemented, failure will occur at instantiation rather
    than dynamically at runtime.

    For example, we enforce children classes to implement `validation_step`
    and `test_step` in addition to just `training_step` as required by the base
    LightningModules. When developing new models there's no strict need for
    `validation` and `test` phases, but for a standard Task we wish to enforce
    this to allow us to collect appropriate metrics.


    Example Usage:

        from pytorch_lightning import LightningModule
        from typing import TypedDict
        if TYPE_CHECKING:
            from torch import Tensor

        class MyTaskOutput(TypedDict):
            predictions: Tensor

        class MyTask(TaskBase[MyTaskOutput], LightningModule):
            def training_step(...):
                ...

            ...
    """

    """ LightningModule Methods. See corresponding methods in pl.LightningModule for documentation. """

    @abstractmethod
    def training_step(
        self, batch: TBatch, batch_idx: int, *args: Any, **kwargs: Any
    ) -> TTrainReturn:
        raise NotImplementedError

    @abstractmethod
    def validation_step(
        self, batch: TBatch, batch_idx: int, *args: Any, **kwargs: Any
    ) -> TTestReturn:
        raise NotImplementedError

    @abstractmethod
    def test_step(
        self, batch: TBatch, batch_idx: int, *args: Any, **kwargs: Any
    ) -> TTestReturn:
        raise NotImplementedError

    @abstractmethod
    def configure_optimizers(
        self,
    ) -> Union[Optimizer, Tuple[Iterable[Optimizer], Iterable[_LRScheduler]]]:
        raise NotImplementedError

    """ Optional methods. Enforce static types for batch. """

    def on_train_batch_start(
        self, batch: TBatch, batch_idx: int, dataloader_idx: int
    ) -> None:
        ...

    def on_train_batch_end(
        self,
        outputs: Iterable[TTrainReturn],
        batch: TBatch,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        ...

    def on_validation_batch_start(
        self, batch: TBatch, batch_idx: int, dataloader_idx: int
    ) -> None:
        ...

    def on_validation_batch_end(
        self,
        outputs: Iterable[TTestReturn],
        batch: TBatch,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        ...

    def on_test_batch_start(
        self, batch: TBatch, batch_idx: int, dataloader_idx: int
    ) -> None:
        ...

    def on_test_batch_end(
        self,
        outputs: Iterable[TTestReturn],
        batch: TBatch,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        ...



# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from dataclasses import dataclass
from typing import Optional

from hydra.core.config_store import ConfigStore
from omegaconf import MISSING
from stl.lightning.conf.trainer import TrainerConf


@dataclass
class ModuleConf:
    pass


@dataclass
class DataModuleConf:
    pass


@dataclass
class TrainAppConf:
    module: ModuleConf = MISSING
    datamodule: Optional[DataModuleConf] = None
    trainer: TrainerConf = MISSING


# pyre-fixme[5]: Global expression must be annotated.
cs = ConfigStore.instance()
cs.store(group="schema/trainer", name="trainer", node=TrainerConf, package="trainer")



def register_components() -> None:
    """
    Calls register_components() for all subfolders so we can register
    subcomponents to Hydra's ConfigStore.
    """
    import pytorch.recipes.rec.datamodules.criteo_datamodule  # noqa
    import pytorch.recipes.rec.datamodules.random_rec_datamodule  # noqa
    import pytorch.recipes.rec.modules.sparsenn_classification  # noqa
    import pytorch.recipes.rec.sparsenn_binary_classification_train_app  # noqa



#!/usr/bin/env python3


def register_components() -> None:
    """
    Imports all python files in the folder to trigger the
    code to register them to Hydra's ConfigStore.
    """

    import pytorch.recipes.text.module.doc_classification  # noqa
    import pytorch.recipes.text.module.doc_classification_demo  # noqa



#!/usr/bin/env python3

import testslide
import torch
from pytorch.recipes.rec.datamodules.random_rec_datamodule import (
    RandomRecDataModule,
)


class TestRandomRecDataModule(testslide.TestCase):
    def test_manual_seed_generator(self) -> None:
        dm1 = RandomRecDataModule(manual_seed=353434)
        iterator1 = iter(dm1.init_loader)
        dm2 = RandomRecDataModule(manual_seed=353434)
        iterator2 = iter(dm2.init_loader)

        for _ in range(10):
            batch1 = next(iterator1)
            batch2 = next(iterator2)
            self.assertTrue(torch.equal(batch1.dense_features, batch2.dense_features))
            self.assertTrue(
                torch.equal(
                    batch1.sparse_features.values(), batch2.sparse_features.values()
                )
            )
            self.assertTrue(
                torch.equal(
                    batch1.sparse_features.offsets(), batch2.sparse_features.offsets()
                )
            )
            self.assertTrue(torch.equal(batch1.labels, batch2.labels))

    def test_no_manual_seed_generator(self) -> None:
        dm1 = RandomRecDataModule()
        iterator1 = iter(dm1.init_loader)
        dm2 = RandomRecDataModule()
        iterator2 = iter(dm2.init_loader)

        for _ in range(10):
            batch1 = next(iterator1)
            batch2 = next(iterator2)
            self.assertFalse(torch.equal(batch1.dense_features, batch2.dense_features))
            self.assertFalse(
                torch.equal(
                    batch1.sparse_features.values(), batch2.sparse_features.values()
                )
            )
            # offsets not random
            self.assertTrue(
                torch.equal(
                    batch1.sparse_features.offsets(), batch2.sparse_features.offsets()
                )
            )



#!/usr/bin/env python3

import argparse
import sys
from typing import List

import pytorch_lightning as pl
from pytorch.recipes.rec.datamodules.criteo_datamodule import CriteoDataModule
from pytorch.recipes.rec.modules.dynamic_sparsenn_classification import (
    SparseNNBinaryClassification,
)
from stl.lightning.callbacks.model_checkpoint import ModelCheckpoint
from stl.lightning.loggers import ManifoldTensorBoardLogger
from torchrec.fb.experimental.dynamic_embedding_configs import DynamicEmbeddingBagConfig
from torchrec.fb.experimental.dynamic_embedding_modules import (
    DynamicEmbeddingBagCollection,
)


def parse_args(argv: List[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="torchrec ML app")
    parser.add_argument(
        "--epochs", type=int, default=3, help="number of epochs to train"
    )
    parser.add_argument(
        "--batch_size", type=int, default=32, help="batch size to use for training"
    )
    parser.add_argument("--load_path", type=str, help="checkpoint path to load from")
    parser.add_argument(
        "--output_path",
        type=str,
        help="path to place checkpoints and model outputs",
        required=True,
    )

    return parser.parse_args(argv)


def main(argv: List[str]) -> None:
    args = parse_args(argv)

    datamodule = CriteoDataModule(num_days=22, batch_size=3, num_days_test=2)
    num_embeddings = datamodule.num_embeddings
    keys = datamodule.keys
    embedding_dim = 64

    eb_configs = [
        DynamicEmbeddingBagConfig(
            name=f"t_{feature_name}",
            embedding_dim=embedding_dim,
            max_embedding_size=num_embeddings,
            feature_names=[feature_name],
        )
        for feature_name in keys
    ]
    model = SparseNNBinaryClassification(
        DynamicEmbeddingBagCollection(tables=eb_configs)
    )

    checkpoint_callback = ModelCheckpoint(
        save_top_k=-1,
        has_user_data=False,
        ttl_days=1,
        monitor=None,
    )
    if args.load_path:
        print(f"loading checkpoint: {args.load_path}...")
        model.load_from_checkpoint(checkpoint_path=args.load_path)

    logger = ManifoldTensorBoardLogger()

    trainer = pl.Trainer(
        logger=logger,
        max_epochs=args.epochs,
        callbacks=[checkpoint_callback],
        limit_train_batches=100,
        limit_val_batches=100,
        limit_test_batches=100,
        log_every_n_steps=1,
        # TODO: find a better way to deal with lazy modules in lightning w/o disabling weights_summary
        weights_summary=None,
    )

    trainer.fit(model, datamodule=datamodule)
    trainer.test(model, datamodule=datamodule)


if __name__ == "__main__":
    main(sys.argv[1:])



from dataclasses import dataclass

import torch
from torchrec.sparse.jagged_tensor import KeyedJaggedTensor


@dataclass
class Batch:
    dense_features: torch.Tensor
    sparse_features: KeyedJaggedTensor
    labels: torch.Tensor

    def to(self, device: torch.device, non_blocking: bool = False) -> "Batch":
        return Batch(
            dense_features=self.dense_features.to(
                device=device, non_blocking=non_blocking
            ),
            sparse_features=self.sparse_features.to(
                device=device, non_blocking=non_blocking
            ),
            labels=self.labels.to(device=device, non_blocking=non_blocking),
        )

    def record_stream(self, stream: torch.cuda.streams.Stream) -> None:
        self.dense_features.record_stream(stream)
        self.sparse_features.record_stream(stream)
        self.labels.record_stream(stream)



#!/usr/bin/env python3

import random
from dataclasses import dataclass, field
from typing import Union, Any, Dict, List, Mapping, Optional, Tuple

import hydra
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from hydra.core.config_store import ConfigStore
from omegaconf import MISSING, OmegaConf
from pytorch.recipes.multimodal.module import (
    MultimodalModuleConf,
    EmbeddingConf,
    EncoderConf,
    FusionConf,
)
from pytorch.recipes.multimodal.utils import (
    build_encoders,
    build_fusion,
    validate_multimodal_config,
    MultimodalClassificationMixupScheme,
)
from pytorch.recipes.utils.config_utils import get_class_name_str
from pytorch.recipes.utils.mixup_utils import MixupParams, MixupUtil
from pytorch_lightning.utilities import rank_zero_info
from torchmetrics import MetricCollection
from torchmetrics.classification import Recall


@dataclass
class ClassifierConf:
    input_fusion_name: str = MISSING
    task_name: str = "single_task"
    in_features: int = MISSING
    out_features: int = MISSING
    # pyre-ignore
    loss: Any = MISSING


class MultimodalClassificationModel(torch.nn.Module):
    def __init__(
        self,
        encoders: Mapping[str, torch.nn.Module],
        fusions: Mapping[str, torch.nn.Module],
        fusion_inputs: Mapping[str, List[str]],
        classifiers: Mapping[str, torch.nn.Linear],
        classifier_input_fusion: Mapping[str, str],
        mixup_params: Optional[MixupParams] = None,
    ) -> None:
        super().__init__()
        self.encoders = encoders
        self.fusions = fusions
        self.fusion_inputs = fusion_inputs
        self.classifiers = classifiers
        self.classifier_input_fusion = classifier_input_fusion
        self.nonlinearity: torch.nn.Module = torch.nn.Softmax()
        self.mixup_params = mixup_params

    def forward(self, input: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        output = {}

        for encoder in self.encoders.values():
            encoder_output = encoder(input)
            output.update(encoder_output)
        for fusion_name, fusion in self.fusions.items():
            fusion_input = {key: output[key] for key in self.fusion_inputs[fusion_name]}
            output.update(fusion(fusion_input))

        model_out = {}
        for task_name, classifier in self.classifiers.items():
            model_out[task_name] = classifier(
                output[self.classifier_input_fusion[task_name]]
            )
            if not self.training:
                model_out[task_name] = self.nonlinearity(model_out[task_name])

        return model_out

    def _mixup_input(
        self,
        input: Dict[str, torch.Tensor],
        mixup_scheme: MultimodalClassificationMixupScheme,
        mixup_util: MixupUtil,
    ) -> Dict[str, torch.Tensor]:
        """Mixes up the input if required"""
        input_copy = {}
        if mixup_scheme == MultimodalClassificationMixupScheme.INPUT:
            for channel_name in self.encoders:
                input_copy[channel_name] = mixup_util.mixup(input[channel_name])
        else:
            input_copy = input
        return input_copy

    def _mixup_encoder_forward(
        self,
        input: Dict[str, torch.Tensor],
        mixup_scheme: MultimodalClassificationMixupScheme,
        mixup_util: MixupUtil,
    ) -> Dict[str, torch.Tensor]:
        """Does a forward pass on the encoders and mixes up the
        encoder output (if required)
        """
        output = {}
        for modality, encoder in self.encoders.items():
            encoder_output = encoder(input)
            if mixup_scheme == MultimodalClassificationMixupScheme.AFTERENCODING:
                encoder_output[modality] = mixup_util.mixup(encoder_output[modality])
            output.update(encoder_output)
        return output

    def _mixup_fusion_forward(
        self,
        input: Dict[str, torch.Tensor],
        mixup_scheme: MultimodalClassificationMixupScheme,
        mixup_util: MixupUtil,
    ) -> Dict[str, torch.Tensor]:
        """Dose a forward pass on the fusion layers and
        mixes up the output of the fusion layers (if required)
        """
        for fusion_name, fusion in self.fusions.items():
            fusion_input = {key: input[key] for key in self.fusion_inputs[fusion_name]}
            fusion_output = fusion(fusion_input)
            if mixup_scheme == MultimodalClassificationMixupScheme.AFTERFUSION:
                for modality in fusion_output:
                    fusion_output[modality] = mixup_util.mixup(fusion_output[modality])
            input.update(fusion_output)
        return input

    def _mixup_classifier_forward(
        self,
        input: Dict[str, torch.Tensor],
        mixup_scheme: MultimodalClassificationMixupScheme,
        mixup_util: MixupUtil,
    ) -> Dict[str, torch.Tensor]:
        """Does a forward pass on the classifiers and mixup the output of
        classifiers (if required)"""
        model_out = {}
        for task_name, classifier in self.classifiers.items():
            model_out[task_name] = classifier(
                input[self.classifier_input_fusion[task_name]]
            )
            if mixup_scheme == MultimodalClassificationMixupScheme.AFTERCLASSIFIER:
                model_out[task_name] = mixup_util.mixup(model_out[task_name])
            if not self.training:
                model_out[task_name] = self.nonlinearity(model_out[task_name])
        return model_out

    def mixup_forward(
        self,
        input: Dict[str, torch.Tensor],
        mixup_util: MixupUtil,
    ) -> Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]]:
        # pyre-ignore[16]
        mixup_scheme = self.mixup_params.scheme
        if mixup_scheme == MultimodalClassificationMixupScheme.RANDOM:
            mixup_scheme = random.choice(
                MultimodalClassificationMixupScheme.get_all_except_RANDOM()
            )

        input_copy = self._mixup_input(input, mixup_scheme, mixup_util)
        # Forward passes
        encoder_out = self._mixup_encoder_forward(input_copy, mixup_scheme, mixup_util)
        fusion_out = self._mixup_fusion_forward(encoder_out, mixup_scheme, mixup_util)
        model_out = self._mixup_classifier_forward(fusion_out, mixup_scheme, mixup_util)

        # mixup the lables
        mixed_target = {}
        for task in model_out:
            mixed_target[task] = mixup_util.mixup_labels(input[task])

        return model_out, mixed_target


class MultimodalClassification(pl.LightningModule):
    def __init__(
        self,
        encoders: Mapping[str, torch.nn.Module],
        fusions: Mapping[str, torch.nn.Module],
        fusion_inputs: Mapping[str, List[str]],
        classifiers: Mapping[str, torch.nn.Linear],
        classifier_input_fusion: Mapping[str, str],
        losses: Mapping[str, torch.nn.Module],
        optimizer: torch.optim.Optimizer,
        freeze_prefixes: Optional[List[str]] = None,
        mixup_params: Optional[MixupParams] = None,
    ) -> None:
        """
        Args:
            encoders: Mapping of channel name to encoder that should be applied to each
                input modality.
            fusions: Mapping of unique fusion identifier to fusion module
            fusion_inputs: Mapping of fusion identifier to the list of input channel names
                that should be passed to the fusion
            classifiers: Mapping of unique task names to Linear layers for classification
            classifier_input_fusion: Mapping of task name to the input fusion name for the
                task
            losses: Mapping of task name to the loss module for the task
            optimizer: torch.optim.Optimizer object for training
            freeze_prefixes: list of prefixes. Any weights that start with these will be
                frozen before training.
            mixup_params: Parameters that specify how and where to apply mixup
        """
        super().__init__()

        self.model = MultimodalClassificationModel(
            encoders=encoders,
            fusions=fusions,
            fusion_inputs=fusion_inputs,
            classifiers=classifiers,
            classifier_input_fusion=classifier_input_fusion,
            mixup_params=mixup_params,
        )
        self.mixup_params = mixup_params
        self.losses = losses
        self.optimizer = optimizer
        self.freeze_prefixes: List[str] = freeze_prefixes or []
        # pyre-fixme[4]: Attribute must be annotated.
        self.example_input_array = None

        self.build_metrics()

    def get_tasks_model(self, task_list: List[str]) -> MultimodalClassificationModel:
        """
        Return MultimodalClassificationModel for specfied tasks
        """
        encoders = torch.nn.ModuleDict()
        fusions = torch.nn.ModuleDict()
        fusion_inputs = {}
        classifiers = torch.nn.ModuleDict()
        classifier_input_fusion = {}

        fusion_names = set()
        encoder_names = set()
        for task in task_list:
            classifiers[task] = self.model.classifiers[task]
            classifier_input_fusion[task] = self.model.classifier_input_fusion[task]
            fusion_names.add(classifier_input_fusion[task])

        for fusion_name in fusion_names:
            fusions[fusion_name] = self.model.fusions[fusion_name]
            fusion_inputs[fusion_name] = self.model.fusion_inputs[fusion_name]
            encoder_names.update(fusion_inputs[fusion_name])

        for encoder_name in encoder_names:
            encoders[encoder_name] = self.model.encoders[encoder_name]

        return MultimodalClassificationModel(
            encoders=encoders,
            fusions=fusions,
            fusion_inputs=fusion_inputs,
            classifiers=classifiers,
            classifier_input_fusion=classifier_input_fusion,
        )

    def build_metrics(self) -> None:
        for phase in ("train", "val", "test"):
            task_metrics = torch.nn.ModuleDict()
            for task in self.model.classifiers:
                task_metrics[task] = MetricCollection(
                    {"recall": Recall(self.model.classifiers[task].out_features)}
                )
            setattr(self, f"{phase}_metrics", task_metrics)

    def setup(self, stage: Optional[str] = None) -> None:
        if stage != "fit" or not self.freeze_prefixes:
            return
        trainable_params = []
        frozen_params = []
        for name, param in self.named_parameters():
            should_be_frozen = any(
                name.startswith(prefix) for prefix in self.freeze_prefixes
            )
            if should_be_frozen:
                frozen_params.append(name)
                param.requires_grad = False
            else:
                trainable_params.append(name)

        rank_zero_info(
            f"Frozen params: {frozen_params}\nTrainable params: {trainable_params}"
        )

    # Do not remove.  Needed to maintain torchscriptability, see https://fburl.com/wiki/b5f028m0
    # pyre-ignore[14] - `forward` overrides method defined in `pl.core.lightning.LightningModule`
    # inconsistently. Could not find parameter `Variable(unknown)` [`Keywords(unknown)`] in overriding signature.
    def forward(self, input: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        return self.model(input)

    def compute_loss_with_mixup(
        self,
        mixup_util: MixupUtil,
        model_out: Dict[str, torch.Tensor],
        original_target: Dict[str, torch.Tensor],
        mixedup_target: Dict[str, torch.Tensor],
    ) -> torch.Tensor:
        loss = []
        for task_name in model_out:
            if task_name in original_target:
                criterion = self.losses[task_name]
                loss.append(
                    mixup_util.compute_loss(
                        criterion,
                        model_out[task_name],
                        original_target[task_name],
                        mixedup_target[task_name],
                    )
                )

        return sum(loss) / len(loss)

    def compute_loss(
        self, model_out: Dict[str, torch.Tensor], target: Dict[str, torch.Tensor]
    ) -> torch.Tensor:
        loss = []
        for task_name in model_out:
            if task_name in target:
                loss.append(
                    self.losses[task_name](model_out[task_name], target[task_name])
                )

        return sum(loss) / len(loss)

    def update_and_compute_metrics(
        self,
        model_out: Dict[str, torch.Tensor],
        target: Dict[str, torch.Tensor],
        phase_type: str,
    ) -> Dict[str, Any]:
        metrics = getattr(self, f"{phase_type}_metrics")
        metric_dict = {}
        for task_name, metrics in metrics.items():
            if task_name not in target:
                continue
            preds = model_out[task_name]
            if phase_type == "train":
                preds = F.softmax(preds)
            value = metrics(preds, target[task_name])
            for metric_name in value:
                metric_dict[f"{phase_type}/{task_name}/{metric_name}"] = value[
                    metric_name
                ]
        return metric_dict

    def reset_metrics(self, phase_type: str) -> None:
        metrics = getattr(self, f"{phase_type}_metrics")
        for _, metric in metrics.items():
            metric.reset()

    # pyre-fixme[2]: Parameter must be annotated.
    def calculate_epoch_end_loss(self, outputs, phase_type: str) -> None:
        losses = [output["loss"] for output in outputs]
        avg_loss = torch.mean(torch.stack(losses))
        self.log(f"Losses/{phase_type}_loss", avg_loss)

    def mixup_forward(
        self, input: Dict[str, torch.Tensor], mixup_util: MixupUtil
    ) -> Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]]:
        return self.model.mixup_forward(input, mixup_util)

    @staticmethod
    def _get_mixup_util(batch: Dict[str, torch.Tensor]) -> MixupUtil:
        batch_size = batch[list(batch)[0]].shape[0]
        return MixupUtil(batch_size)

    def _step(
        self,
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        phase_type: str,
    ) -> Dict[str, torch.Tensor]:
        if phase_type == "train" and self.mixup_params:
            mixup_util = self._get_mixup_util(batch)
            model_out, mixed_up_target = self.mixup_forward(batch, mixup_util)
            loss = self.compute_loss_with_mixup(
                mixup_util, model_out, batch, mixed_up_target
            )
        else:
            model_out = self.forward(batch)
            loss = self.compute_loss(model_out, batch)
        self.log(f"{phase_type}_loss", loss)
        metric_dict = self.update_and_compute_metrics(
            model_out, batch, phase_type=phase_type
        )
        self.log_dict(metric_dict)
        return {"loss": loss}

    def training_step(
        self,
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> Dict[str, torch.Tensor]:
        return self._step(batch, batch_idx, "train")

    def validation_step(
        self,
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> Dict[str, torch.Tensor]:
        return self._step(batch, batch_idx, "val")

    def test_step(
        self,
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> Dict[str, torch.Tensor]:
        return self._step(batch, batch_idx, "test")

    def training_epoch_end(
        self, outputs: List[Union[Dict[str, Any], torch.Tensor]]
    ) -> None:
        self.calculate_epoch_end_loss(outputs, "train")
        self.reset_metrics("train")

    def validation_epoch_end(
        self, outputs: List[Union[Dict[str, Any], torch.Tensor]]
    ) -> None:
        self.calculate_epoch_end_loss(outputs, "val")
        self.reset_metrics("val")

    def test_epoch_end(
        self, outputs: List[Union[Dict[str, Any], torch.Tensor]]
    ) -> None:
        self.calculate_epoch_end_loss(outputs, "test")
        self.reset_metrics("test")

    def configure_optimizers(self) -> torch.optim.Optimizer:
        return self.optimizer


@dataclass
class MultimodalClassificationModuleConf(MultimodalModuleConf):
    _target_: str = get_class_name_str(MultimodalClassification)
    classifier: List[ClassifierConf] = field(default_factory=list)


cs: ConfigStore = ConfigStore.instance()


cs.store(
    group="schema/module",
    name="multimodal_classification",
    node=MultimodalClassificationModuleConf,
    package="module",
)


def validate_multimodal_classification_config(
    embeddings_conf: List[EmbeddingConf],
    encoders_conf: List[EncoderConf],
    fusion_conf: List[FusionConf],
    classifier_conf: List[ClassifierConf],
) -> None:
    """
    Validate input configurations for MultimodalClassification
    """

    # validate common multimodal components here
    validate_multimodal_config(
        embeddings_conf=embeddings_conf,
        encoders_conf=encoders_conf,
        fusion_conf=fusion_conf,
    )

    # check if classifier input fusion names exist
    fusion_names = [fusion.fusion_name for fusion in fusion_conf]
    for classifier in classifier_conf:
        if classifier.input_fusion_name not in fusion_names:
            raise AssertionError(
                f"{classifier.input_fusion_name} not found in fusion configuration"
            )


def convert_hydra_conf_to_module(
    module_conf: MultimodalClassificationModuleConf, validate_config: bool = True
) -> MultimodalClassification:
    """
    Args:
        module_conf: Config to build MultimodalClassification module. Following
            keys are expected to be in the module_conf.
            embeddings: word embeddings to be used with corresponding channel name
            encoders: This is a list of model configs that should be applied to each
                input modality. Each model config contains 3 keys:
                channel_name: input channel_name that will be processed via this encoder
                encoder_args: constructor parameters for the encoder
                dict_input: boolean value to control whether the encoder expects a
                    dict input or a tensor input.
            fusion: This is a list of fusion modules which define how the outputs of the
                encoders are fused. Each fusion config contains three keys:
                fusion_name: A unique name to identify this fusion module
                input_encoder_names: The encoder outputs that should be passed to this
                    fusion module
                fusion_args: constructor parameters for the fusion module
            classifier: This is a list of parameters to define the linear laers, one
                for each task. Each classifier config contains 5 keys:
                input_fusion_name: The identifier for the fusion module whose output
                    is passed to the classifier
                task_name: A unique name for the classifier. For a single task, this could be "single_task"
                in_features: The input dimension for the linear layer
                out_features: The output classes for the task
                loss: The loss function for the task
            optim: the optimizer to be used
            freeze_prefixes: list of prefixes. Any weights that start with these will be
                frozen before training.
        validate_config: boolean to control whether the input config should be validated
            before building.
    """

    # build embeddings and encoders if provided
    # both embeddings and encoders have channel_name to identify correspondence.

    if validate_config:
        validate_multimodal_classification_config(
            embeddings_conf=module_conf.embeddings,
            encoders_conf=module_conf.encoders,
            fusion_conf=module_conf.fusion,
            classifier_conf=module_conf.classifier,
        )
    encoders = build_encoders(
        encoders_conf=module_conf.encoders, embeddings_conf=module_conf.embeddings
    )

    # build fusion modules
    fusions, fusion_inputs = build_fusion(
        fusion_conf=module_conf.fusion, encoders=encoders
    )

    # build classifiers
    classifiers = torch.nn.ModuleDict()
    classifier_input_fusion = {}
    losses = torch.nn.ModuleDict()

    for classifier in module_conf.classifier:
        in_features = (
            classifier.in_features or fusions[classifier.input_fusion_name].output_dim
        )
        classifiers[classifier.task_name] = torch.nn.Linear(
            in_features, classifier.out_features
        )
        classifier_input_fusion[classifier.task_name] = classifier.input_fusion_name

        losses[classifier.task_name] = hydra.utils.instantiate(classifier.loss)

    parameters = (
        # pyre-ignore
        list(encoders.parameters())
        + list(fusions.parameters())
        + list(classifiers.parameters())
    )
    optimizer = hydra.utils.instantiate(module_conf.optim, parameters)
    mixup_params = None
    if module_conf.mixup:
        mixup_params = OmegaConf.to_container(module_conf.mixup)
        # pyre-ignore
        mixup_params["scheme"] = MultimodalClassificationMixupScheme[
            # pyre-ignore
            mixup_params["scheme"]
        ]
        # pyre-ignore
        mixup_params = MixupParams(**mixup_params)

    return MultimodalClassification(
        encoders=encoders,
        fusions=fusions,
        fusion_inputs=fusion_inputs,
        classifiers=classifiers,
        classifier_input_fusion=classifier_input_fusion,
        losses=losses,
        optimizer=optimizer,
        freeze_prefixes=module_conf.freeze_prefixes,
        mixup_params=mixup_params,
    )



#!/usr/bin/env python3

from dataclasses import dataclass, field
from typing import Any, Dict, List, Mapping, Optional

import hydra
import pytorch_lightning as pl
import torch
from hydra.core.config_store import ConfigStore
from omegaconf import MISSING, OmegaConf
from pytorch.recipes.multimodal.module import MultimodalModuleConf
from pytorch.recipes.multimodal.utils import (
    build_encoders,
    build_fusion,
)
from pytorch.recipes.utils.config_utils import get_class_name_str
from pytorch_lightning.utilities import rank_zero_info


@dataclass
class TowerConf:
    input_fusion_name: List[str] = MISSING
    task_name: str = "single_task"
    # pyre-ignore
    output_fusion: Any = MISSING
    # pyre-ignore
    loss: Any = MISSING


class MultimodalTwoTowerModel(torch.nn.Module):
    def __init__(
        self,
        encoders: Mapping[str, torch.nn.Module],
        fusions: Mapping[str, torch.nn.Module],
        fusion_inputs: Mapping[str, List[str]],
        towers: Mapping[str, torch.nn.Module],
        tower_input_fusion: Mapping[str, List[str]],
    ) -> None:
        super().__init__()
        self.encoders = encoders
        self.fusions = fusions
        self.fusion_inputs = fusion_inputs
        self.towers = towers
        self.tower_input_fusion = tower_input_fusion

    def forward(self, input: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        output = {}
        for encoder in self.encoders.values():
            encoder_output = encoder(input)
            output.update(encoder_output)
        for fusion_name, fusion in self.fusions.items():
            fusion_input = {key: output[key] for key in self.fusion_inputs[fusion_name]}
            output.update(fusion(fusion_input))

        model_out = {}
        for task_name, tower in self.towers.items():
            tower_input = [
                output[fusion_name]
                for fusion_name in self.tower_input_fusion[task_name]
            ]
            model_out[task_name] = tower(tower_input)

        return model_out


class MultimodalTwoTower(pl.LightningModule):
    def __init__(
        self,
        encoders: Mapping[str, torch.nn.Module],
        fusions: Mapping[str, torch.nn.Module],
        fusion_inputs: Mapping[str, List[str]],
        towers: Mapping[str, torch.nn.Module],
        tower_input_fusion: Mapping[str, List[str]],
        losses: Mapping[str, torch.nn.Module],
        optimizer: torch.optim.Optimizer,
        freeze_prefixes: Optional[List[str]] = None,
    ) -> None:

        super().__init__()
        self.model = MultimodalTwoTowerModel(
            encoders=encoders,
            fusions=fusions,
            fusion_inputs=fusion_inputs,
            towers=towers,
            tower_input_fusion=tower_input_fusion,
        )
        self.losses = losses
        self.optimizer = optimizer
        self.freeze_prefixes: List[str] = freeze_prefixes or []
        # pyre-fixme[4]: Attribute must be annotated.
        self.example_input_array = None

    def get_tasks_model(self, task_list: List[str]) -> MultimodalTwoTowerModel:
        """
        Return MultimodalClassificationModel for specfied tasks
        """
        encoders = torch.nn.ModuleDict()
        fusions = torch.nn.ModuleDict()
        fusion_inputs = {}
        towers = torch.nn.ModuleDict()
        tower_input_fusion = {}

        fusion_names = set()
        encoder_names = set()
        for task in task_list:
            towers[task] = self.model.towers[task]
            tower_input_fusion[task] = self.model.tower_input_fusion[task]
            fusion_names.update(tower_input_fusion[task])

        for fusion_name in fusion_names:
            fusions[fusion_name] = self.model.fusions[fusion_name]
            fusion_inputs[fusion_name] = self.model.fusion_inputs[fusion_name]
            encoder_names.update(fusion_inputs[fusion_name])

        for encoder_name in encoder_names:
            encoders[encoder_name] = self.model.encoders[encoder_name]

        return MultimodalTwoTowerModel(
            encoders=encoders,
            fusions=fusions,
            fusion_inputs=fusion_inputs,
            towers=towers,
            tower_input_fusion=tower_input_fusion,
        )

    def setup(self, stage: Optional[str] = None) -> None:
        if stage != "fit" or not self.freeze_prefixes:
            return
        frozen_params = []
        for name, param in self.named_parameters():
            should_be_frozen = any(
                name.startswith(prefix) for prefix in self.freeze_prefixes
            )
            if should_be_frozen:
                frozen_params.append(name)
                param.requires_grad = False

        rank_zero_info(f"Frozen params: {frozen_params}")

    # Do not remove.  Needed to maintain torchscriptability, see https://fburl.com/wiki/b5f028m0
    # pyre-ignore[14] - `forward` overrides method defined in `pl.core.lightning.LightningModule`
    # inconsistently. Could not find parameter `Variable(unknown)` [`Keywords(unknown)`] in overriding signature.
    def forward(self, input: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        return self.model(input)

    def compute_loss(
        self, model_out: Dict[str, torch.Tensor], target: Dict[str, torch.Tensor]
    ) -> torch.Tensor:
        loss = []
        for task_name in model_out:
            loss.append(self.losses[task_name](model_out[task_name], target[task_name]))

        return sum(loss) / len(loss)

    def _step(
        self,
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        phase_type: str,
    ) -> Dict[str, torch.Tensor]:
        model_out = self.forward(batch)
        loss = self.compute_loss(model_out, batch)
        self.log(f"{phase_type}_loss", loss, on_epoch=True)
        return {"loss": loss}

    def training_step(
        self,
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> Dict[str, torch.Tensor]:
        return self._step(batch, batch_idx, "train")

    def validation_step(
        self,
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> Dict[str, torch.Tensor]:
        return self._step(batch, batch_idx, "val")

    def test_step(
        self,
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> Dict[str, torch.Tensor]:
        return self._step(batch, batch_idx, "test")

    def configure_optimizers(self) -> torch.optim.Optimizer:
        return self.optimizer


@dataclass
class MultimodalTwoTowerModuleConf(MultimodalModuleConf):
    _target_: str = get_class_name_str(MultimodalTwoTower)
    towers: List[TowerConf] = field(default_factory=list)


cs: ConfigStore = ConfigStore.instance()


cs.store(
    group="schema/module",
    name="multimodal_two_tower",
    node=MultimodalTwoTowerModuleConf,
    package="module",
)


def convert_hydra_conf_to_module(
    module_conf: MultimodalTwoTowerModuleConf,
) -> MultimodalTwoTower:
    # build embeddings and encoders if provided
    # both embeddings and encoders have channel_name to identify correspondence.
    encoders = build_encoders(
        encoders_conf=module_conf.encoders, embeddings_conf=module_conf.embeddings
    )

    # build fusion modules
    fusions, fusion_inputs = build_fusion(
        fusion_conf=module_conf.fusion, encoders=encoders
    )

    # build towers
    tower_input_fusion = {}
    towers = torch.nn.ModuleDict()
    losses = torch.nn.ModuleDict()
    for tower in module_conf.towers:
        towers[tower.task_name] = hydra.utils.instantiate(tower.output_fusion)

        tower_input_fusion[tower.task_name] = OmegaConf.to_container(
            tower.input_fusion_name
        )

        losses[tower.task_name] = hydra.utils.instantiate(tower.loss)

    parameters = (
        # pyre-ignore
        list(encoders.parameters())
        + list(fusions.parameters())
        + list(towers.parameters())
    )
    optimizer = hydra.utils.instantiate(module_conf.optim, parameters)

    return MultimodalTwoTower(
        encoders=encoders,
        fusions=fusions,
        fusion_inputs=fusion_inputs,
        towers=towers,
        tower_input_fusion=tower_input_fusion,
        losses=losses,
        optimizer=optimizer,
        freeze_prefixes=module_conf.freeze_prefixes,
    )



#!/usr/bin/env python3

import testslide
from pytorch.recipes.rec.datamodules.criteo_datamodule import CriteoDataModule


class TestCriteoDataModule(testslide.TestCase):
    def test_none_stage(self) -> None:
        dm = CriteoDataModule(num_days=1, batch_size=3, num_days_test=1, num_workers=0)
        dm.setup()

        train_batch = next(iter(dm.train_dataloader()))
        self.assertEqual(train_batch.dense_features.size(), (3, 13))
        kjt = train_batch.sparse_features
        self.assertEqual(kjt.lengths().size(), (26 * 3,))
        self.assertEqual(kjt.keys(), dm.keys)

        test_batch = next(iter(dm.test_dataloader()))
        self.assertEqual(test_batch.dense_features.size(), (3, 13))
        kjt = test_batch.sparse_features
        self.assertEqual(kjt.lengths().size(), (26 * 3,))
        self.assertEqual(kjt.keys(), dm.keys)

    def test_fit_stage(self) -> None:
        dm = CriteoDataModule(num_days=1, batch_size=3, num_days_test=1, num_workers=0)
        dm.setup(stage="fit")

        train_batch = next(iter(dm.train_dataloader()))
        self.assertEqual(train_batch.dense_features.size(), (3, 13))
        kjt = train_batch.sparse_features
        self.assertEqual(kjt.lengths().size(), (26 * 3,))
        self.assertEqual(kjt.keys(), dm.keys)

        with self.assertRaises(AssertionError):
            # only train/val dataloaders are set up
            dm.test_dataloader()

    def test_test_stage(self) -> None:
        dm = CriteoDataModule(num_days=1, batch_size=3, num_days_test=1, num_workers=0)
        dm.setup(stage="test")

        with self.assertRaises(AssertionError):
            # only test dataloader is set up
            dm.train_dataloader()

        test_batch = next(iter(dm.test_dataloader()))
        self.assertEqual(test_batch.dense_features.size(), (3, 13))
        kjt = test_batch.sparse_features
        self.assertEqual(kjt.lengths().size(), (26 * 3,))
        self.assertEqual(kjt.keys(), dm.keys)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from dataclasses import dataclass
from typing import Dict

import torch
from hydra.core.config_store import ConfigStore
from pytorch.recipes.core.conf import DataModuleConf
from pytorch.recipes.utils.config_utils import get_class_name_str
from pytorch_lightning import LightningDataModule
from torch.utils.data.dataset import Dataset


class MultimodalDataset(Dataset[Dict[str, torch.Tensor]]):
    def __init__(self, only_floats: bool = False) -> None:
        self.len = 1024
        if only_floats:
            self.modality_1_dataset: torch.Tensor = torch.randn(self.len, 128)
        else:
            self.modality_1_dataset: torch.Tensor = torch.randint(
                0, 10, size=(self.len, 128)
            )
        self.modality_2_dataset: torch.Tensor = torch.randn(self.len, 256)
        self.labels: torch.Tensor = torch.randint(0, 2, size=(self.len,))

    def __getitem__(self, index: int) -> Dict[str, torch.Tensor]:
        return {
            "modality_1": self.modality_1_dataset[index],
            "modality_2": self.modality_2_dataset[index],
            "single_task": self.labels[index],
        }

    def __len__(self) -> int:
        return self.len


class TestMultimodalDataModule(LightningDataModule):

    only_floats: bool

    def __init__(self, only_floats: bool = False) -> None:
        super().__init__()
        self.only_floats = only_floats

    def train_dataloader(self) -> torch.utils.data.DataLoader:
        return torch.utils.data.DataLoader(
            MultimodalDataset(only_floats=self.only_floats), batch_size=128
        )

    def val_dataloader(self) -> torch.utils.data.DataLoader:
        return torch.utils.data.DataLoader(
            MultimodalDataset(only_floats=self.only_floats), batch_size=128
        )

    def test_dataloader(self) -> torch.utils.data.DataLoader:
        return torch.utils.data.DataLoader(
            MultimodalDataset(only_floats=self.only_floats), batch_size=128
        )


@dataclass
class TestMultimodalDataModuleConf(DataModuleConf):
    _target_: str = get_class_name_str(TestMultimodalDataModule)
    only_floats: bool = False


cs: ConfigStore = ConfigStore.instance()

cs.store(
    group="schema/datamodule",
    name="test_multimodal_datamodule",
    node=TestMultimodalDataModuleConf,
    package="datamodule",
)



# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from dataclasses import dataclass, field
from typing import List, Any

# @manual "//github/facebookresearch/hydra:hydra"
from hydra.core.config_store import ConfigStore
from pytorch.recipes.core.base_train_app import BaseTrainApp
from pytorch.recipes.core.conf.base_config import BaseTrainAppConf
from pytorch.recipes.multimodal.datamodule.test_multimodal_datamodule import (
    TestMultimodalDataModuleConf,
)
from pytorch.recipes.multimodal.module.multimodal_classification import (
    MultimodalClassificationModuleConf,
    convert_hydra_conf_to_module as convert_hydra_conf_to_module_classification,
    MultimodalClassification,
)
from pytorch.recipes.multimodal.module.multimodal_two_tower import (
    MultimodalTwoTowerModuleConf,
    convert_hydra_conf_to_module as convert_hydra_conf_to_module_two_tower,
    MultimodalTwoTower,
)
from pytorch.recipes.utils.config_utils import get_class_name_str
from stl.lightning.conf.trainer import TrainerConf


multimodal_classification_defaults = [
    # Module
    "_self_",
    {"schema/module": "multimodal_classification"},
    # Module doesn't take any direct fields. No dataclass required.
    {"module/embeddings": "test_embedding"},
    {"module/encoders": "test_encoders"},
    {"module/fusion": "test_fusion"},
    {"module/classifier": "test_classifier"},
    {"module/optim": "adam"},
    # Data
    {"schema/datamodule": "test_multimodal_datamodule"},
    # Trainer
    {"trainer": "cpu"},
]

multimodal_two_tower_defaults = [
    # Module
    "_self_",
    {"schema/module": "multimodal_two_tower"},
    # Module doesn't take any direct fields. No dataclass required.
    {"module/embeddings": "test_embedding"},
    {"module/encoders": "test_encoders"},
    {"module/fusion": "test_fusion"},
    {"module/towers": "test_towers"},
    {"module/optim": "adam"},
    # Data
    {"schema/datamodule": "test_multimodal_datamodule"},
    # Trainer
    {"trainer": "cpu"},
]


class MultimodalClassificationTrainApp(BaseTrainApp):
    def __init__(
        self,
        module: MultimodalClassificationModuleConf,
        trainer: TrainerConf,
        datamodule: TestMultimodalDataModuleConf,
    ) -> None:
        super().__init__(module, trainer, datamodule)

    def get_lightning_module(self) -> MultimodalClassification:
        # pyre-ignore
        module = convert_hydra_conf_to_module_classification(self.module_conf)
        return module


@dataclass
class MultimodalClassificationTrainAppConf(BaseTrainAppConf):
    _target_: str = get_class_name_str(MultimodalClassificationTrainApp)
    # pyre-fixme[4]: Attribute annotation cannot contain `Any`.
    defaults: List[Any] = field(
        default_factory=lambda: multimodal_classification_defaults
    )
    module: MultimodalClassificationModuleConf = MultimodalClassificationModuleConf()


class MultimodalTwoTowerTrainApp(BaseTrainApp):
    def __init__(
        self,
        module: MultimodalTwoTowerModuleConf,
        trainer: TrainerConf,
        datamodule: TestMultimodalDataModuleConf,
    ) -> None:
        super().__init__(module, trainer, datamodule)

    def get_lightning_module(self) -> MultimodalTwoTower:
        # pyre-ignore
        module = convert_hydra_conf_to_module_two_tower(self.module_conf)
        return module


@dataclass
class MultimodalTwoTowerTrainAppConf(BaseTrainAppConf):
    _target_: str = get_class_name_str(MultimodalTwoTowerTrainApp)
    # pyre-fixme[4]: Attribute annotation cannot contain `Any`.
    defaults: List[Any] = field(default_factory=lambda: multimodal_two_tower_defaults)
    module: MultimodalTwoTowerModuleConf = MultimodalTwoTowerModuleConf()


cs: ConfigStore = ConfigStore.instance()

cs.store(
    name="multimodal_classification_app", node=MultimodalClassificationTrainAppConf
)
cs.store(name="multimodal_two_tower_app", node=MultimodalTwoTowerTrainAppConf)



#!/usr/bin/env python3

# pyre-strict

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

import hydra
from hydra.core.config_store import ConfigStore
from pytext.fb.utils.manifold_utils import register_manifold_handler
from pytorch.recipes.core.base_train_app import BaseTrainApp
from pytorch.recipes.core.conf import TrainAppConf
from pytorch.recipes.text.datamodule.doc_classification import (
    DocClassificationDataModuleConf,
)
from pytorch.recipes.text.module.doc_classification_demo import (
    DocClassificationModuleConf,
)
from pytorch.recipes.utils.config_utils import get_class_name_str
from pytorch_lightning.core.datamodule import LightningDataModule
from stl.lightning.conf.trainer import TrainerConf


class DocClassificationTrainApp(BaseTrainApp):
    """
    This app is used to launch the doc classification training / testing.
    """

    module_conf: DocClassificationModuleConf

    def __init__(
        self,
        module: DocClassificationModuleConf,
        trainer: TrainerConf,
        datamodule: DocClassificationDataModuleConf,
    ) -> None:
        register_manifold_handler()
        super().__init__(module, trainer, datamodule)

    def get_data_module(self) -> Optional[LightningDataModule]:
        transform_cfg = self.module_conf.transform
        transform = hydra.utils.instantiate(transform_cfg)
        datamodule = hydra.utils.instantiate(self.datamodule_conf, transform=transform)
        return datamodule


defaults: List[Union[str, Dict[str, str]]] = [
    "_self_",
    # Module
    {"schema/module": "doc_classification"},
    # Module doesn't take any direct fields. No dataclass required.
    # Model
    {"module/model": "dummy_doc_model"},
    # Transform
    {"module/transform": "dummy_xlmr_transform"},
    # Optim
    {"module/optim": "adamw"},
    # Data
    {"schema/datamodule": "doc_classification"},
    {"datamodule": "dummy_sst2_manifold"},
    # Trainer
    {"trainer": "cpu_1_epoch"},
]


@dataclass
class DocClassificationTrainAppConf(TrainAppConf):
    _target_: str = get_class_name_str(DocClassificationTrainApp)
    # pyre-ignore: Missing attribute annotation [4]: Attribute `defaults`
    # of class `DocClassificationTrainAppConf` is used as type `List[Dict[str, str]]`
    # and must have a type that does not contain `Any`.
    defaults: List[Any] = field(default_factory=lambda: defaults)


cs: ConfigStore = ConfigStore.instance()

cs.store(name="doc_classification_app", node=DocClassificationTrainAppConf)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

# pyre-ignore-all-errors

from typing import Any, Dict

import torch
from apex import amp
from pytorch_lightning.utilities import AMPType, DeviceType


class CheckpointMixin(object):
    """Mixin to enable resuming from checkpoint
    Currently, resuming from checkpointing requires a hack in `on_pretrain_routine_end`.
    TODO: T98699770 @stevenliu remove this class after the official fix landed in Lightning
    Usage:
        MyTask(CheckpointMixin, LightningModule):
            ...

    Note: CheckpointMixin must be added ahead of LightningModule. For FSDP, it's
    required to add attribute `enable_configure_sharded_model` to Task and set it to True
    """

    def on_pretrain_routine_end(self) -> None:
        if self.trainer is None:
            return

        if self.trainer.resume_from_checkpoint is None:
            return

        # Before reconnecting, as we already restored optimizer states, lr scheduler, amp states
        #  we store it temporarily and after reconnecting, we will load it.
        restored_checkpoint = self._get_restored_trainer_states()

        # Reconnecting model, configure, and pre dispatch.
        self.trainer.accelerator.connect(self)
        self.trainer.accelerator.setup_environment()
        if getattr(self, "enable_configure_sharded_model", False):
            self.trainer._call_configure_sharded_model(self)
        self.trainer.optimizers = []
        self.trainer.accelerator.setup(self.trainer, self)
        self.trainer.accelerator.pre_dispatch(self.trainer)

        # Restore the optimizers, lr scheduler, amp states for re-connected and configured model
        self._restore_trainer_states(restored_checkpoint)

    def _get_restored_trainer_states(self) -> Dict[str, Any]:
        restored_checkpoint = {}
        optimizer_states = []
        for _, optimizer in enumerate(self.trainer.optimizers):
            # Rely on accelerator to dump optimizer state
            optimizer_state = self.trainer.accelerator.optimizer_state(optimizer)
            optimizer_states.append(optimizer_state)
        restored_checkpoint["optimizer_states"] = optimizer_states
        # dump lr schedulers
        lr_schedulers = []
        for scheduler in self.trainer.lr_schedulers:
            lr_schedulers.append(scheduler["scheduler"].state_dict())
        restored_checkpoint["lr_schedulers"] = lr_schedulers
        # dump amp scaling
        if (
            self.trainer.amp_backend == AMPType.NATIVE
            and self.trainer._device_type != DeviceType.TPU
            and self.trainer.scaler is not None
        ):
            restored_checkpoint[
                "native_amp_scaling_state"
            ] = self.trainer.scaler.state_dict()
        elif self.trainer.amp_backend == AMPType.APEX:
            restored_checkpoint["amp_scaling_state"] = amp.state_dict()
        return restored_checkpoint

    def _restore_trainer_states(self, checkpoint: Dict[str, Any]) -> None:
        # restore the optimizers
        optimizer_states = checkpoint["optimizer_states"]
        for optimizer, opt_state in zip(self.trainer.optimizers, optimizer_states):
            optimizer.load_state_dict(opt_state)

            # move optimizer to GPU 1 weight at a time
            # avoids OOM
            if self.trainer.root_gpu is not None:
                for state in optimizer.state.values():
                    for k, v in state.items():
                        if isinstance(v, torch.Tensor):
                            state[k] = v.cuda(self.trainer.root_gpu)
        # restore the lr schedulers
        lr_schedulers = checkpoint["lr_schedulers"]
        for scheduler, lrs_state in zip(self.trainer.lr_schedulers, lr_schedulers):
            scheduler["scheduler"].load_state_dict(lrs_state)
        # restore amp scaling
        if (
            self.trainer.amp_backend == AMPType.NATIVE
            and "native_amp_scaling_state" in checkpoint
        ):
            self.trainer.scaler.load_state_dict(checkpoint["native_amp_scaling_state"])
        elif (
            self.trainer.amp_backend == AMPType.APEX
            and "amp_scaling_state" in checkpoint
        ):
            amp.load_state_dict(checkpoint["amp_scaling_state"])



import glob
import os
import re
import shutil
import subprocess
from typing import List

from setuptools import setup, find_packages
import distutils.command.clean  # isort:skip

_PACKAGE_NAME: str = "torchrecipes"
_VERSION_FILE: str = "version.py"
_README: str = "README.md"
_REQUIREMENTS: str = "requirements.txt"
_DEV_REQUIREMENTS: str = "dev-requirements.txt"
_GITIGNORE: str = ".gitignore"


def get_version() -> str:
    """Retrieves the version of the library."""
    if version := os.getenv("BUILD_VERSION"):
        return version
    cwd = os.path.dirname(os.path.abspath(__file__))
    version_file_path = os.path.join("pytorch", "recipes", _VERSION_FILE)
    version_regex = r"__version__: str = ['\"]([^'\"]*)['\"]"
    with open(version_file_path, "r") as f:
        search = re.search(version_regex, f.read(), re.M)
    assert search
    version = search.group(1)

    try:
        sha = (
            subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=cwd)
            .decode("ascii")
            .strip()
        )
        version += "+" + sha[:7]
    except Exception:
        pass
    return version


def get_long_description() -> str:
    """Fetch project description as Markdown."""
    with open(_README, mode="r") as f:
        return f.read()


def get_requirements() -> List[str]:
    """Fetch requirements."""
    with open(_REQUIREMENTS, mode="r") as f:
        return f.readlines()


def get_dev_requirements() -> List[str]:
    """Fetch requirements for library development."""
    with open(_DEV_REQUIREMENTS, mode="r") as f:
        return f.readlines()


class clean(distutils.command.clean.clean):
    def run(self) -> None:
        with open(_GITIGNORE, "r") as f:
            ignores = f.readlines()
            for wildcard in filter(None, ignores):
                for filename in glob.glob(wildcard):
                    try:
                        os.remove(filename)
                    except OSError:
                        shutil.rmtree(filename, ignore_errors=True)

        # It's an old-style class in Python 2.7...
        distutils.command.clean.clean.run(self)


if __name__ == "__main__":
    version: str = get_version()
    print("Building wheel {}-{}".format(_PACKAGE_NAME, version))

    setup(
        # Metadata
        name=_PACKAGE_NAME,
        version=version,
        author="PyTorch Ecosystem Foundations Team",
        author_email="luispe@fb.com",
        description="Prototype of training recipes for PyTorch",
        long_description=get_long_description(),
        long_description_content_type="text/markdown",
        url="https://github.com/facebookexternal/stl_tasks",
        license="BSD-3",
        keywords=["pytorch", "machine learning"],
        python_requires=">=3.8",
        install_requires=get_requirements(),
        include_package_data=True,
        # Package info
        packages=find_packages(),
        cmdclass={
            "clean": clean,
        },
        extras_require={"dev": get_dev_requirements()},
        # PyPI package information.
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: BSD License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.8",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
        ],
    )



#!/usr/bin/env python3

import importlib
import logging
import os
from enum import auto, unique, Enum
from typing import Union, Optional

import hydra
from omegaconf import OmegaConf
from pytorch.recipes.core.base_train_app import BaseTrainApp
from pytorch.recipes.core.base_train_app import TrainOutput
from pytorch.recipes.core.conf.base_config import BaseTrainAppConf
from pytorch_lightning.utilities.types import _EVALUATE_OUTPUT, _PREDICT_OUTPUT
from torch.distributed.elastic.multiprocessing import errors

logger: logging.Logger = logging.getLogger(__name__)

# Set default value of these environment variables in
# fbcode/torchx/components/fb/stl_apps.py

# Your TrainApp's hydra conf module. We need to import this module before calling hydra
CONFIG_MODULE = "CONFIG_MODULE"
# Which mode your App will run in.
#   - prod: (Default) train + test, return test result
#   - train: train only
#   - test: test only
#   - predict: train + predict
MODE = "MODE"


@unique
class Mode(Enum):
    PROD = auto()
    TRAIN = auto()
    TEST = auto()
    PREDICT = auto()


def _get_mode() -> Mode:
    """Fetch operating environment."""
    mode_key = os.getenv(MODE, "").upper()
    try:
        return Mode[mode_key]
    except KeyError:
        logger.warning("Unknown MODE, run train and test by default")
        return Mode.PROD


def run_in_certain_mode(
    app: BaseTrainApp,
) -> Union[TrainOutput, _EVALUATE_OUTPUT, Optional[_PREDICT_OUTPUT]]:
    mode = _get_mode()
    if mode == Mode.TRAIN:
        logger.info("MODE set to train, run train only.")
        return app.train()
    elif mode == Mode.TEST:
        logger.info("MODE set to test, run test only.")
        return app.test()
    elif mode == Mode.PREDICT:
        logger.info("MODE set to predict, run train and precit.")
        app.train()
        return app.predict()
    else:
        # By default, run train and test
        app.train()
        return app.test()


@hydra.main()
def run_with_hydra(
    cfg: BaseTrainAppConf,
) -> Union[TrainOutput, _EVALUATE_OUTPUT, Optional[_PREDICT_OUTPUT]]:
    logger.info(OmegaConf.to_yaml(cfg))
    app = hydra.utils.instantiate(cfg, _recursive_=False)
    return run_in_certain_mode(app)


# pyre-ignore[56]: Decorator is not defined in typeshed_internal.
@errors.record
def main() -> None:
    config_module = os.getenv(CONFIG_MODULE)
    logger.info(f"CONFIG_MODULE: {config_module}")
    # only needed for apps that use hydra config
    if config_module:
        importlib.import_module(config_module)
        run_with_hydra()
    else:
        # TODO: T93277666 add entry point for non-hydra apps
        raise NotImplementedError


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from dataclasses import dataclass
from enum import Enum

import numpy as np
import torch


class MixupScheme(Enum):
    """Mixup scheme: Where to perform mixup within a model."""


@dataclass
class MixupParams:
    """
    alpha: float. Mixup ratio. Recommended values from (0.125, 0.25, 0.5, 1.0, 2.0, 4.0, 8.0)
    scheme: MixupScheme. The locations to perform mixup within a model.
    More details about the parameters and mixup can be found in FAIM WIKI:
    https://fburl.com/wiki/3f0qh0zr
    """

    alpha: float
    scheme: MixupScheme


class MixupUtil:
    @staticmethod
    def _get_lambda(alpha: float = 1.0) -> float:
        # Sample from a beta distribution
        # The result is used for linear interpolation
        if alpha > 0.0:
            lam = np.random.beta(alpha, alpha)
        else:
            lam = 1.0
        return lam

    def __init__(self, batch_size: int) -> None:
        self.indices: torch.Tensor = torch.randperm(batch_size)
        self.lam: float = self._get_lambda()

    def mixup(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.lam + x[self.indices] * (1 - self.lam)

    def compute_loss(
        self,
        criterion: torch.nn.Module,
        pred: torch.Tensor,
        original_target: torch.Tensor,
        mixed_target: torch.Tensor,
    ) -> float:
        return self.lam * criterion(pred, original_target) + (1 - self.lam) * criterion(
            pred, mixed_target
        )

    def mixup_labels(self, x: torch.Tensor) -> torch.Tensor:
        return x[self.indices]



#!/usr/bin/env python3

import os
import unittest
import uuid

import pytorch_lightning as pl
import torch
from pytorch.recipes.rec.accelerators.torchrec import (
    TorchrecAccelerator,
)
from pytorch.recipes.rec.datamodules.random_rec_datamodule import RandomRecDataModule
from pytorch.recipes.rec.modules.sharded_sparsenn_classification import (
    ShardedSparseNNBinaryClassification,
)
from torch.distributed.launcher.api import elastic_launch, LaunchConfig
from torchrec import EmbeddingBagCollection
from torchrec.modules.embedding_configs import EmbeddingBagConfig
from torchrec.tests.utils import skip_if_asan


class TestShardedSparseNNClassification(unittest.TestCase):
    @classmethod
    def _run_trainer(cls) -> None:
        torch.manual_seed(int(os.environ["RANK"]))
        num_embeddings = 100
        embedding_dim = 12

        eb1_config = EmbeddingBagConfig(
            name="t1",
            embedding_dim=embedding_dim,
            num_embeddings=num_embeddings,
            feature_names=["f1", "f3"],
        )
        eb2_config = EmbeddingBagConfig(
            name="t2",
            embedding_dim=embedding_dim,
            num_embeddings=num_embeddings,
            feature_names=["f2"],
        )
        eb_configs = [eb1_config, eb2_config]

        lit_models = []
        datamodules = []
        for _ in range(2):
            ebc = EmbeddingBagCollection(tables=eb_configs, device=torch.device("meta"))
            lit_model = ShardedSparseNNBinaryClassification(
                ebc,
                over_embedding_dim=5,
                hidden_layer_size=20,
            )

            datamodule = RandomRecDataModule(manual_seed=564733621)
            init_batch = next(iter(datamodule.init_loader)).to(lit_model.device)
            with torch.no_grad():
                lit_model(init_batch.dense_features, init_batch.sparse_features)
                lit_model.model.init_data_parallel()

            lit_models.append(lit_model)
            datamodules.append(datamodule)
        lit_model1, lit_model2 = lit_models
        dm1, dm2 = datamodules

        # Load m1 state dicts into m2
        lit_model2.model.load_state_dict(lit_model1.model.state_dict(), strict=False)
        optim1 = lit_model1.configure_optimizers()
        optim2 = lit_model2.configure_optimizers()
        optim2.load_state_dict(optim1.state_dict())

        # train model 1 using lightning
        trainer = pl.Trainer(
            max_epochs=1,
            limit_train_batches=5,
            limit_val_batches=5,
            limit_test_batches=5,
            accelerator=TorchrecAccelerator(),
            weights_summary=None,
            logger=None,
        )
        trainer.fit(lit_model1, datamodule=dm1)

        # train model 2 manually
        train_dataiterator = iter(dm2.train_dataloader())
        for _ in range(5):
            batch = next(train_dataiterator).to(lit_model2.device)
            optim2.zero_grad()
            logits = lit_model2.model(batch.dense_features, batch.sparse_features)
            loss = lit_model2.loss_fn(logits.squeeze(), batch.labels.float())
            loss.backward()
            optim2.step()

        # assert parameters equal
        sd1 = lit_model1.model.state_dict()
        for name, value in lit_model2.model.state_dict().items():
            assert torch.equal(sd1[name], value)

        # assert model evaluation equal
        test_dataiterator = iter(dm2.test_dataloader())
        with torch.no_grad():
            for _ in range(10):
                batch = next(test_dataiterator).to(lit_model2.device)
                assert torch.equal(
                    lit_model2.model(batch.dense_features, batch.sparse_features),
                    lit_model1.model(batch.dense_features, batch.sparse_features),
                )

    @skip_if_asan
    def test_lit_trainer_equivalent_to_non_lit(self) -> None:
        lc = LaunchConfig(
            min_nodes=1,
            max_nodes=1,
            nproc_per_node=2,
            run_id=str(uuid.uuid4()),
            # TODO T100608035 replace with c10d when fixed
            rdzv_backend="zeus",
            rdzv_endpoint="",
            start_method="spawn",
            monitor_interval=1,
            max_restarts=0,
        )

        elastic_launch(config=lc, entrypoint=self._run_trainer)()



#!/usr/bin/env python3
from dataclasses import dataclass, field
from typing import Any, List, Optional

from omegaconf import MISSING
from pytorch.recipes.core.conf import ModuleConf


@dataclass
class EmbeddingConf:
    channel_name: str = MISSING
    # pyre-ignore
    embedding_args: Any = MISSING


@dataclass
class EncoderConf:
    channel_name: str = MISSING
    # pyre-ignore
    encoder_args: Any = MISSING
    dict_input: bool = True


@dataclass
class FusionConf:
    fusion_name: str = MISSING
    input_encoder_names: List[str] = MISSING
    # pyre-ignore
    fusion_args: Any = MISSING


@dataclass
class MixupConf:
    alpha: float = MISSING
    scheme: str = MISSING


@dataclass
class MultimodalModuleConf(ModuleConf):
    embeddings: List[EmbeddingConf] = field(default_factory=list)
    encoders: List[EncoderConf] = field(default_factory=list)
    fusion: List[FusionConf] = field(default_factory=list)
    # pyre-ignore
    optim: Any = MISSING
    freeze_prefixes: List[str] = field(default_factory=list)
    mixup: Optional[MixupConf] = None


def register_components() -> None:
    """
    Imports all python files in the folder to trigger the
    code to register them to Hydra's ConfigStore.
    """

    import pytorch.recipes.multimodal.module.multimodal_classification  # noqa
    import pytorch.recipes.multimodal.module.multimodal_two_tower  # noqa



# pyre-strict

import pickle
from abc import ABC, abstractmethod

import testslide
from pytorch_lightning import LightningModule


class TaskTestCaseBase(ABC, testslide.TestCase):
    """All Standard Task unit tests should inherit from this class."""

    @abstractmethod
    def get_standard_task(self) -> LightningModule:
        """Subclasses should implement a standard method of retrieving and instance of the Task to test."""
        raise NotImplementedError

    def test_standard_task_is_torchscriptable(self) -> None:
        task = self.get_standard_task()
        _ = task.to_torchscript()

    def test_standard_task_is_pickleable(self) -> None:
        task = self.get_standard_task()
        pickle.dumps(task)
